__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "vivarium_inputs"
__summary__ = "Transformations and artifact building for the vivarium microsimulation project."
__uri__ = "https://github.com/ihmeuw/vivarium_inputs"

__version__ = "1.0.9"

__author__ = "The vivarium developers"
__email__ = "vivarium.dev@gmail.com"

__license__ = "GNU GPLv3"
__copyright__ = f"Copyright 2016-2018 {__author__}"



import pytest

from gbd_mapping import causes, risk_factors, etiologies



@pytest.fixture
def cause_list():
    return [causes.diarrheal_diseases, causes.ischemic_heart_disease, causes.ischemic_stroke,
            causes.hemorrhagic_stroke, causes.tetanus, causes.diabetes_mellitus, causes.all_causes]


@pytest.fixture
def etiology_list():
    return [etiologies.cholera, etiologies.amoebiasis]


@pytest.fixture
def sequela_list():
    return list(causes.diarrheal_diseases.sequelae + causes.ischemic_heart_disease.sequelae
                + causes.ischemic_stroke.sequelae + causes.hemorrhagic_stroke.sequelae
                + causes.hemorrhagic_stroke.sequelae + causes.tetanus.sequelae
                + causes.diabetes_mellitus.sequelae)


@pytest.fixture
def etiology_list():
    return list(causes.diarrheal_diseases.etiologies + causes.lower_respiratory_infections.etiologies)


@pytest.fixture
def risk_list():
    return [r for r in risk_factors]


@pytest.fixture
def locations():
    return ['Bangladesh', 'Ethiopia', 'Kenya', 'China', 'North Korea', 'Nigeria']



from typing import Set
import warnings

from gbd_mapping import causes, risk_factors, sequelae, coverage_gaps, covariates, etiologies
import pandas as pd
import numpy as np

from vivarium_public_health.dataset_manager import EntityKey

from vivarium_inputs import core
from vivarium_inputs.data_artifact.utilities import normalize
from vivarium_inputs.utilities import normalize_for_simulation, get_age_group_midpoint_from_age_group_id
from vivarium_inputs.mapping_extension import healthcare_entities, health_technologies


class DataArtifactError(Exception):
    """Base error raised for issues in data artifact construction."""
    pass


class EntityError(DataArtifactError):
    """Error raised when modeled entities are improperly specified."""
    pass


CAUSE_BY_ID = {c.gbd_id: c for c in causes if c is not None}
RISK_BY_ID = {r.gbd_id: r for r in risk_factors}

AGE_COLS = ['age', 'age_group_start', 'age_group_end']
YEAR_COLS = ['year', 'year_start', 'year_end']

def loader(entity_key: EntityKey, location: str, modeled_causes: Set[str], all_measures: bool=False):
    entity_data = {
        "cause": {
            "mapping": causes,
            "getter": get_cause_data,
            "measures": ["death", "prevalence", "incidence", "cause_specific_mortality",
                         "excess_mortality", "population_attributable_fraction", "remission",
                         "sequelae", "etiologies", "restrictions"]
        },
        "risk_factor": {
            "mapping": risk_factors,
            "getter": get_risk_data,
            "measures": ["affected_causes", "affected_risk_factors", "restrictions", "distribution",
                         "exposure_parameters", "levels", "tmred", "exposure", "exposure_standard_deviation",
                         "relative_risk", "ensemble_weights"],
        },
        "sequela": {
            "mapping": sequelae,
            "getter": get_sequela_data,
            "measures": ["healthstate", "prevalence", "incidence", "disability_weight"],
        },
        "healthcare_entity": {
            "mapping": healthcare_entities,
            "getter": get_healthcare_entity_data,
            "measures": ["cost", "annual_visits"]
        },
        "health_technology": {
            "mapping": health_technologies,
            "getter": get_health_technology_data,
            "measures": ["cost"]
        },
        "coverage_gap": {
            "mapping": coverage_gaps,
            "getter": get_coverage_gap_data,
            "measures": ["affected_causes", "affected_risk_factors", "restrictions", "distribution", "levels",
                         "population_attributable_fraction", "relative_risk", "exposure"]
        },
        "etiology": {
            "mapping": etiologies,
            "getter": get_etiology_data,
            "measures": ["population_attributable_fraction"],
        },
        "population": {
            "mapping": {'': None},
            "getter": get_population_data,
            "measures": ["structure", "age_bins", "theoretical_minimum_risk_life_expectancy"],
        },
        "covariate": {
            "mapping": covariates,
            "getter": get_covariate_data,
            "measures": ["estimate"]
        },
        "subregions": {
            "mapping": {'': None},
            "getter": get_subregion_data,
            "measures": ["sub_region_ids"],
        },
        "dimensions": {
            "mapping": {'': None},
            "getter": get_dimension_data,
            "measures": ["full_space"]
        },
    }
    mapping, getter, measures = entity_data[entity_key.type].values()
    entity = mapping[entity_key.name]
    if not all_measures:
        return getter(entity, entity_key.measure, location, modeled_causes)
    else:
        return _data_generator(entity, measures, location, modeled_causes, getter)


def _data_generator(entity, measures, location, modeled_causes, getter):
    for measure in measures:
        data = getter(entity, measure, location, modeled_causes)
        if data is not None:
            yield measure, data
            del data


def get_cause_data(cause, measure, location, _):
    if measure in ["sequelae", "etiologies", "restrictions"]:
        data = _get_cause_metadata(cause, measure)
    elif measure in ["death", "prevalence", "incidence", "cause_specific_mortality", "excess_mortality"]:
        data = core.get_draws([cause], [measure], [location])
        data = normalize(data)[["location", "sex", "draw", "value"] + AGE_COLS + YEAR_COLS]
    elif measure == "population_attributable_fraction":
        data = _get_cause_population_attributable_fraction(cause, location)
    elif measure == "remission":
        data = _get_cause_remission(cause, location)
    else:
        raise NotImplementedError(f"Unknown measure {measure} for cause {cause.name}")

    return data


def get_risk_data(risk, measure, location, modeled_causes):
    if measure in ["affected_causes", "affected_risk_factors", "restrictions",
                   "distribution", "exposure_parameters", "levels", "tmred"]:
        data = _get_risk_metadata(risk, measure, modeled_causes)
    elif measure == "exposure":
        data = _get_risk_exposure(risk, location)
    elif measure == "exposure_standard_deviation":
        data = _get_risk_exposure_standard_deviation(risk, location)
    elif measure == "relative_risk":
        data = _get_risk_relative_risk(risk, location)
    elif measure == "ensemble_weights":
        data = _get_risk_ensemble_weights(risk)
    else:
        raise NotImplementedError(f"Unknown measure {measure} for risk {risk.name}")
    return data


def get_sequela_data(sequela, measure, location, _):
    if measure == "healthstate":
        data = sequela.healthstate.name
    elif measure in ["incidence", "prevalence"]:
        data = core.get_draws([sequela], [measure], [location]).drop("sequela_id", axis=1)
        data = normalize(data)[["location", "sex", "draw", "value"] + AGE_COLS + YEAR_COLS]
        data["sequela_id"] = sequela.gbd_id
    elif measure == "disability_weight":
        data = core.get_draws([sequela], ["disability_weight"], [location])
        index_columns = [c for c in data.columns if "draw_" not in c]
        draw_columns = [c for c in data.columns if "draw_" in c]
        data = pd.melt(data, id_vars=index_columns, value_vars=draw_columns, var_name="draw")
        data["draw"] = data.draw.str.partition("_")[2].astype(int)
    else:
        raise NotImplementedError(f"Unknown measure {measure} for sequela {sequela.name}")
    return data


def get_healthcare_entity_data(healthcare_entity, measure, location, _):
    if measure == "cost":
        data = core.get_draws([healthcare_entity], ["cost"], [location])
        data = normalize(data)
        data = data.loc[data.sex == 'Male', ["location", "draw", "value"] + YEAR_COLS]
    elif measure == "annual_visits":
        data = core.get_draws([healthcare_entity], ["annual_visits"], [location])
        data = normalize(data)
        data = data[["sex", "value", "draw"] + AGE_COLS + YEAR_COLS]
    else:
        raise NotImplementedError(f"Unknown measure {measure} for healthcare_entity {healthcare_entity.name}")
    return data


def get_health_technology_data(healthcare_technology, measure, location, _):
    if measure == "cost":
        data = core.get_draws([healthcare_technology], ["cost"], [location])
        data = normalize(data)[["location", "draw", "value", "health_technology"] + YEAR_COLS]
    else:
        raise NotImplementedError(f"Unknown measure {measure} for healthcare_entity {healthcare_technology.name}")
    return data


def get_coverage_gap_data(coverage_gap, measure, location, modeled_causes):
    if measure in ["affected_causes", "affected_risk_factors", "restrictions", "distribution", "levels"]:
        data = _get_coverage_gap_metadata(coverage_gap, measure, modeled_causes)
    elif measure == "exposure":
        data = _get_coverage_gap_exposure(coverage_gap, location)
    elif measure == "relative_risk":
        data = _get_coverage_gap_relative_risk(coverage_gap, location)
    elif measure == "population_attributable_fraction":
        data = _get_coverage_gap_population_attributable_fraction(coverage_gap, location)
    else:
        raise NotImplementedError(f"Unknown measure {measure} for coverage_gap {coverage_gap.name}")
    return data


def get_etiology_data(etiology, measure, location, _):
    if measure == "population_attributable_fraction":
        data = core.get_draws([etiology], ["population_attributable_fraction"], [location])
        data = normalize(data)
        data["cause"] = data.cause_id.apply(lambda cause_id: CAUSE_BY_ID[cause_id].name)
        data = data[["location", "cause", "sex", "draw", "value"] + AGE_COLS + YEAR_COLS]
    else:
        raise NotImplementedError(f"Unknown measure {measure} for etiology {etiology.name}")
    return data


def get_population_data(_, measure, location, __):
    if measure == "structure":
        data = core.get_populations([location])
        data = normalize_for_simulation(data)
        data = get_age_group_midpoint_from_age_group_id(data)
    elif measure == "age_bins":
        data = core.get_age_bins()[["age_group_years_start", "age_group_years_end", "age_group_name"]]
        data = data.rename(columns={"age_group_years_start": "age_group_start", "age_group_years_end": "age_group_end"})
    elif measure == "theoretical_minimum_risk_life_expectancy":
        data = core.get_theoretical_minimum_risk_life_expectancy()
    else:
        raise NotImplementedError(f"Unknown measure {measure} for population.")
    return data


def get_covariate_data(covariate, measure, location, _):
    if measure == "estimate":
        data = core.get_covariate_estimates([covariate], [location])
        expected_columns = ["location", "mean_value", "lower_value", "upper_value",
                            "sex_id", "year_id", "age_group_id"]

        if not set(expected_columns).issubset(data.columns):
            raise NotImplementedError(f"Unsupported covariate {covariate.name}. It does not not "
                                       f"contain the columns {expected_columns}.")
        data = data[expected_columns]
        if (data['age_group_id'] == 22).all():
            data = data.drop(['age_group_id'], axis=1)
            warnings.warn(f"Covariate \"{covariate.name}\" contains data for the age group all ages, "
                           "so the age column is being dropped.")
        else:
            data = get_age_group_midpoint_from_age_group_id(data)
        data = normalize_for_simulation(data)
    else:
        raise NotImplementedError((f"Unknown or unsupported measure {measure} for ")
                                  (f"covariate {covariate.name}"))
    return data


def get_subregion_data(_, measure, location, __):
    if measure == "sub_region_ids":
        data = pd.DataFrame(core.get_subregions([location]))
        data = data.melt(var_name="location", value_name="subregion_id")
    else:
        raise NotImplementedError(f"Unknown measure {measure} for subregion data.")
    return data


def get_dimension_data(_, measure, location, __):
    if measure == "full_space":
        age_bins = core.get_age_bins()
        estimation_years = core.get_estimation_years()
        data = [range(min(estimation_years), max(estimation_years) + 1),
                ["Male", "Female"], age_bins.age_group_id, [location]]
        data = pd.MultiIndex.from_product(data, names=["year", "sex", "age_group_id", "location"])
        data = data.to_frame().reset_index(drop=True)
    else:
        raise NotImplementedError(f"Unknown measure {measure} for dimensions")
    return data


##########
# Causes #
##########


def _get_cause_metadata(entity, field):
    if field == "restrictions":
        data = entity.restrictions.to_dict()
    else:  # field in ["sequela", "etiologies"]:
        if entity[field] is not None:
            data = [sub_entity.name for sub_entity in entity[field]]
        else:
            data = None
    return data


def _get_cause_population_attributable_fraction(cause, location):
    if cause.name == "all_causes":
        data = None
    else:
        pafs = core.get_draws([cause], ["population_attributable_fraction"], [location])
        if pafs.empty:
            warnings.warn(f"No Population Attributable Fraction data found for cause '{cause.name}'")
            data = None
        else:
            normalized = []
            for key, group in pafs.groupby(["risk_id"]):
                group = group.drop(["risk_id"], axis=1)
                group = normalize(group)
                if key in RISK_BY_ID:
                    group["risk_factor"] = RISK_BY_ID[key].name
                    dims = ["year", "year_start", "year_end", "sex", "measure", "age", "age_group_start",
                            "age_group_end", "location", "draw", "risk_factor"]
                    normalized.append(group.set_index(dims))
                else:
                    warnings.warn(f"Found risk_id {key} in population attributable fraction data for cause "
                                  f"'{cause.name}' but that risk is missing from the gbd mapping")
            data = pd.concat(normalized).reset_index()
            data = data[["location", "sex", "draw", "value", "risk_factor"] + AGE_COLS + YEAR_COLS]
    return data


def _get_cause_remission(cause, location):
    try:
        result = core.get_draws([cause], ["remission"], [location])
        if not result.empty:
            data = normalize(result)[["location", "sex", "draw", "value"] + AGE_COLS + YEAR_COLS]
        else:
            data = None
    except core.InvalidQueryError:
        data = None

    return data


#########
# Risks #
#########


def _get_risk_metadata(risk, measure, modeled_causes):
    if measure in ["restrictions", "exposure_parameters", "levels", "tmred"]:
        if risk[measure] is not None:
            data = risk[measure].to_dict()
        else:
            data = None
    elif measure == "affected_causes":
        data = [c.name for c in risk.affected_causes if c.name in modeled_causes]
    elif measure == "affected_risk_factors":
        # FIXME: Update mapping to include affected risks (mediation)
        data = []
    else:  # measure == "distribution"
        data = risk[measure]
    return data


def _get_risk_exposure(risk, location):
    exposures = core.get_draws([risk], ["exposure"], [location])
    normalized = []
    for key, group in exposures.groupby(["parameter"]):
        group = group.drop(["parameter"], axis=1)
        group = normalize(group)
        group["parameter"] = key
        dims = ["year", "year_start", "year_end", "sex", "measure", "age", "age_group_start",
                "age_group_end", "location", "draw", "parameter"]
        normalized.append(group.set_index(dims))
    result = pd.concat(normalized).reset_index()
    result = result[["location", "sex", "draw", "value", "parameter"] + AGE_COLS + YEAR_COLS]
    return result


def _get_risk_exposure_standard_deviation(risk, location):
    if risk.exposure_parameters is not None:
        exposure_sds = core.get_draws([risk], ["exposure_standard_deviation"], [location])
        exposure_sds = normalize(exposure_sds)
        data = exposure_sds[["location", "sex", "draw", "value"] + AGE_COLS + YEAR_COLS]
    else:
        data = None
    return data


def _get_risk_relative_risk(risk, location):
    rrs = core.get_draws([risk], ["relative_risk"], [location])
    normalized = []
    for key, group in rrs.groupby(["parameter", "cause_id"]):
        group = group.drop(["cause_id", "parameter"], axis=1)
        group = normalize(group)
        group["parameter"] = key[0]
        group["cause"] = CAUSE_BY_ID[key[1]].name
        dims = ["year", "year_start", "year_end", "sex", "measure", "age", "age_group_start",
                "age_group_end", "location", "draw", "cause", "parameter"]
        normalized.append(group.set_index(dims))
    result = pd.concat(normalized).reset_index()
    result = result[["location", "sex", "draw", "value", "parameter", "cause"] + AGE_COLS + YEAR_COLS]
    return result


def _get_risk_ensemble_weights(risk):
    if risk.distribution == "ensemble":
        weights = core.get_ensemble_weights([risk])
        weights = weights.drop(["location_id", "risk_id"], axis=1)
        weights = normalize_for_simulation(weights)
        data = get_age_group_midpoint_from_age_group_id(weights)
    else:
        data = None
    return data


#################
# Coverage Gaps #
#################


def _get_coverage_gap_metadata(coverage_gap, measure, modeled_causes):
    if measure in ["restrictions", "levels"]:
        data = coverage_gap[measure].to_dict()
    elif measure == "affected_causes":
        data = [c.name for c in coverage_gap.affected_causes if c.name in modeled_causes]
    elif measure == "affected_risk_factors":
        data = [r.name for r in coverage_gap.affected_risk_factors]
    else:  # measure == "distribution"
        data = coverage_gap[measure]
    return data


def _get_coverage_gap_exposure(coverage_gap, location):
    exposures = core.get_draws([coverage_gap], ["exposure"], [location])
    normalized = []
    for key, group in exposures.groupby(["parameter"]):
        group = group.drop(["parameter"], axis=1)
        group = normalize(group)
        group["parameter"] = key
        dims = ["year", "year_start", "year_end", "sex", "measure", "age", "age_group_start",
                "age_group_end", "location", "draw", "parameter"]
        normalized.append(group.set_index(dims))
    result = pd.concat(normalized).reset_index()
    result = result[["location", "sex", "draw", "value", "parameter"] + AGE_COLS + YEAR_COLS]
    return result


def _get_coverage_gap_relative_risk(coverage_gap, location):
    data = core.get_draws([coverage_gap], ["relative_risk"], [location])
    if data.empty:
        data = None
    else:
        data = _handle_coverage_gap_rr_paf(data)
        data = data[['location', 'cause', 'risk_factor', 'sex', 'draw', 'value', 'parameter'] + AGE_COLS + YEAR_COLS]
    return data


def _get_coverage_gap_population_attributable_fraction(coverage_gap, location):
    data = core.get_draws([coverage_gap], ["population_attributable_fraction"], [location])
    if data.empty:
        data = None
    else:
        data = _handle_coverage_gap_rr_paf(data)
        data = data[["location", "cause", "risk_factor", "sex", "draw", "value",] + AGE_COLS + YEAR_COLS]
    return data


def _handle_coverage_gap_rr_paf(data):
    data = normalize(data)
    data = data.rename(columns={'cause_id': 'cause', 'risk_id': 'risk_factor'})
    if data['cause'].dropna().unique().size > 0:
        for cid in data['cause'].dropna().unique():
            data['cause'] = data['cause'].apply(lambda c: CAUSE_BY_ID[c].name if c == cid else c)
    if data['risk_factor'].dropna().unique().size > 0:
        for rid in data['risk_factor'].dropna().unique():
            if rid == 318:  # measles
                data.loc[data['risk_factor'] == rid, 'risk_factor'] = 'low_measles_vaccine_coverage_first_dose'
            else:
                data['risk_factor'] = data['risk_factor'].apply(lambda r: RISK_BY_ID[r].name if r == rid else r)
    return data



from datetime import datetime
import logging
from typing import Collection, Any
import pandas as pd

from vivarium_public_health.dataset_manager import EntityKey, Artifact, hdf, get_location_term, filter_data
from vivarium_public_health.disease import DiseaseModel

from vivarium_inputs.data_artifact.loaders import loader


_log = logging.getLogger(__name__)


class ArtifactBuilder:

    def setup(self, builder):
        path = builder.configuration.input_data.artifact_path
        append = builder.configuration.input_data.append_to_artifact
        hdf.touch(path, append)
        draw = builder.configuration.input_data.input_draw_number

        self.location = builder.configuration.input_data.location
        self.artifact = Artifact(path, filter_terms=[f'draw == {draw}', get_location_term(self.location)])
        self.modeled_causes = {c.cause for c in builder.components.get_components(DiseaseModel)}
        self.processed_entities = set()
        self.start_time = datetime.now()

        self.load("dimensions.full_space")

        builder.event.register_listener('post_setup', self.end_processing)

    def load(self, entity_key: str, keep_age_group_edges=True, **__) -> Any:
        entity_key = EntityKey(entity_key)
        if entity_key not in self.artifact:
            self.process(entity_key)
        data = self.artifact.load(entity_key)
        return filter_data(data, keep_age_group_edges, **__) if isinstance(data, pd.DataFrame) else data

    def end_processing(self, event) -> None:
        _log.debug(f"Data loading took at most {datetime.now() - self.start_time} seconds")

    def process(self, entity_key: EntityKey) -> None:
        if entity_key not in self.processed_entities:
            _worker(entity_key, self.location, self.modeled_causes, self.artifact)
            self.processed_entities.add(entity_key)


def _worker(entity_key: EntityKey, location: str, modeled_causes: Collection[str], artifact: Artifact) -> None:
    artifact.write(entity_key, loader(entity_key, location, modeled_causes, all_measures=False))



from typing import Any

import pandas as pd
from vivarium_public_health.dataset_manager import EntityKey, filter_data
from vivarium_public_health.disease import DiseaseModel

from vivarium_inputs.data_artifact.loaders import loader


class ArtifactPassthrough:

    def setup(self, builder):
        self.modeled_causes = builder.components.get_components(DiseaseModel)
        self.location = builder.configuration.input_data.location
        draw = builder.configuration.input_data.input_draw_number

        self.base_filter = {'draw': draw,
                            'location': [self.location, 'Global']}

    def load(self, entity_key: str, keep_age_group_edges: bool=True, **column_filters: str) -> Any:
        entity_key = EntityKey(entity_key)
        data = loader(entity_key, self.location, self.modeled_causes)

        if isinstance(data, pd.DataFrame):
            for key, val in self.base_filter.items():
                if key in data.columns:
                    column_filters[key] = val
            data = filter_data(data, keep_age_group_edges, **column_filters)
        return data



import hashlib

import numpy as np
import pandas as pd
import pytest

from gbd_mapping.id import reiid
from gbd_mapping.cause import causes
from gbd_mapping.risk import risk_factors

from vivarium_inputs import core


def test_get_ids_for_inconsistent_entities(cause_list, sequela_list):
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure(cause_list + sequela_list, 'test')


def test_get_ids_for_death(cause_list, sequela_list):
    ids = core._get_ids_for_measure(cause_list, 'death')
    assert set(ids) == {c.gbd_id for c in cause_list}
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure(sequela_list, 'death')


def test_get_ids_for_remission(sequela_list):
    cause_list = [causes.diarrheal_diseases, causes.tetanus, causes.diabetes_mellitus]
    ids = core._get_ids_for_measure(cause_list, 'remission')
    assert set(ids) == {c.dismod_id for c in cause_list}
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure(sequela_list, 'remission')
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure([causes.age_related_and_other_hearing_loss], 'remission')
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure([causes.ischemic_heart_disease], 'remission')


def test_get_ids_for_prevalence(cause_list, sequela_list, etiology_list):
    ids = core._get_ids_for_measure(cause_list, 'prevalence')
    assert set(ids) == {c.gbd_id for c in cause_list}
    ids = core._get_ids_for_measure(sequela_list, 'prevalence')
    assert set(ids) == {s.gbd_id for s in sequela_list}
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure(etiology_list, 'prevalence')


def test_get_ids_for_incidence(cause_list, sequela_list, etiology_list):
    ids = core._get_ids_for_measure(cause_list, 'incidence')
    assert set(ids) == {c.gbd_id for c in cause_list}
    ids = core._get_ids_for_measure(sequela_list, 'incidence')
    assert set(ids) == {s.gbd_id for s in sequela_list}
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure(etiology_list, 'incidence')


def test_get_ids_for_exposure(cause_list, risk_list):
    ids = core._get_ids_for_measure(risk_list, 'exposure')
    assert set(ids) == {r.gbd_id for r in risk_list}
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure(cause_list, 'exposure')


def test_get_ids_for_relative_risk(cause_list, risk_list):
    ids = core._get_ids_for_measure(risk_list, 'relative_risk')
    assert set(ids) == {r.gbd_id for r in risk_list}
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure(cause_list, 'relative_risk')


def test_get_ids_for_population_attributable_fraction(cause_list, risk_list):
    ids = core._get_ids_for_measure(cause_list, 'population_attributable_fraction')
    assert set(ids) == {r.gbd_id for r in cause_list}
    with pytest.raises(core.InvalidQueryError):
        core._get_ids_for_measure(risk_list, 'population_attributable_fraction')


# TODO there's a bunch of repeated code in the next three functions but I'm not sure what the general form should be yet
@pytest.fixture
def mock_rrs(mocker):
    rrs_mock = mocker.patch("vivarium_inputs.core._get_relative_risk")
    rr_map = {}

    def rr_builder(risks, locations):
        rrs = []
        for risk in risks:
            for cause in risk.affected_causes:
                if (risk.gbd_id, cause.gbd_id) in rr_map:
                    current_rr = rr_map[(risk.gbd_id, cause.gbd_id)]
                else:
                    current_rr = int(hashlib.md5(str((risk.gbd_id, cause.gbd_id)).encode()).hexdigest(), 16)
                    current_rr /= 2**128
                    rr_map[(risk.gbd_id, cause.gbd_id)] = current_rr+1
                age_groups = [10, 11, 12]
                years = [1990, 1995]
                sexes = [1, 2]
                idx = pd.MultiIndex.from_product(
                    [[cause.gbd_id],   [risk.gbd_id], age_groups,     years,     sexes,    locations,  ["continuous"]],
                    names=["cause_id", "risk_id",     "age_group_id", "year_id", "sex_id", "location_id", "parameter"])
                rrs.append(pd.DataFrame({f"draw_{i}": current_rr for i in range(1000)}, index=idx).reset_index())
        return pd.concat(rrs)
    rrs_mock.side_effect = rr_builder
    return rrs_mock, rr_map


@pytest.fixture
def mock_pafs(mocker, cause_list):
    risk_cause_pafs = {}

    def mock_pafs(entity_ids, location_ids):
        pafs = []
        for gbd_id in entity_ids:
            if isinstance(gbd_id, reiid):
                rids = [gbd_id]
                # TODO This assumes that all non-cause entities are diarrhea etiologies
                cids = [causes.diarrheal_diseases.gbd_id]
            else:
                rids = {r.gbd_id for r in risk_factors if gbd_id in [cc.gbd_id for cc in r.affected_causes]}
                cids = [gbd_id]
            for c in cids:
                for r in rids:
                    if risk_cause_pafs:
                        if (r, c) in risk_cause_pafs:
                            current_paf = risk_cause_pafs[(r, c)]
                        else:
                            current_paf = int(hashlib.md5(str((r, c)).encode()).hexdigest(), 16)
                            current_paf /= 2**128
                            risk_cause_pafs[(r, c)] = current_paf
                    else:
                        current_paf = 0.001
                        risk_cause_pafs[(r, c)] = current_paf

                    age_groups = [10, 11, 12]
                    years = [1990, 1995]
                    sexes = [1, 2]
                    idx = pd.MultiIndex.from_product(
                        [[c],               [r],      age_groups,     years,     sexes,    [3],          location_ids],
                        names=["cause_id", "rei_id", "age_group_id", "year_id", "sex_id", "measure_id", "location_id"])
                    pafs.append(pd.DataFrame({f"draw_{i}":current_paf for i in range(1000)}, index=idx).reset_index())
        return pd.concat(pafs)

    gbd_mock = mocker.patch("vivarium_inputs.core.gbd")
    gbd_mock.get_pafs.side_effect = mock_pafs
    return gbd_mock, risk_cause_pafs


@pytest.fixture
def mock_exposures(mocker):
    exposures_mock = mocker.patch("vivarium_inputs.core._get_exposure")
    exposure_map = {}

    def exposure_builder(risks, locations):
        exposures = []
        for risk in risks:
            if risk.gbd_id in exposure_map:
                current_exposure = exposure_map[risk.gbd_id]
            else:
                current_exposure = int(hashlib.md5(str(risk.gbd_id).encode()).hexdigest(), 16)
                current_exposure /= 2**128
                current_exposure *= 100
                exposure_map[risk.gbd_id] = current_exposure
            age_groups = [10, 11, 12]
            years = [1990, 1995]
            sexes = [1, 2]
            idx = pd.MultiIndex.from_product(
                [[risk.gbd_id],   age_groups,     years,     sexes,    locations,  ["continuous"]],
                names=["risk_id", "age_group_id", "year_id", "sex_id", "location_id", "parameter"])
            exposures.append(pd.DataFrame({f"draw_{i}": current_exposure
                                           for i in range(1000)}, index=idx).reset_index())
        return pd.concat(exposures)

    exposures_mock.side_effect = exposure_builder
    return exposures_mock, exposure_map


@pytest.mark.skip("Cluster")
def test__compute_paf_for_special_cases(mock_rrs, mock_exposures, locations):
    _, rrs = mock_rrs
    _, exposures = mock_exposures

    # TODO: This list is canonically specified as a constant inside _get_population_attributable_fraction
    # where it isn't really accessible for tests. Should probably clean that up.
    special_risks = [risk_factors.unsafe_water_source]

    location_ids = [core.get_location_ids_by_name()[name] for name in locations]
    for risk in special_risks:
        for cause in risk.affected_causes:
            paf = core._compute_paf_for_special_cases(cause, risk, location_ids)
            assert cause.gbd_id in paf.cause_id.values
            assert risk.gbd_id in paf.risk_id.values
            e = exposures[risk.gbd_id]
            rr = rrs[(risk.gbd_id, cause.gbd_id)]
            true_paf = (rr*e - 1) / (rr*e)
            assert all(paf[[f"draw_{i}" for i in range(1000)]] == true_paf)


@pytest.fixture(params=["cause", "etiology"])
def cause_like_entities(request, cause_list, etiology_list):
    if request.param == "cause":
        return cause_list
    elif request.param == "etiology":
        return etiology_list

@pytest.mark.skip("Cluster")
def test_get_draws_bad_args(cause_list, risk_list, locations):
    with pytest.raises(core.InvalidQueryError):
        core.get_draws(cause_list + risk_list, ['test'], locations)

    for measure in ['death', 'remission', 'prevalence', 'incidence', 'population_attributable_fraction']:
        with pytest.raises(core.InvalidQueryError):
            core.get_draws(risk_list, [measure], locations)

    for measure in ['exposure', 'relative_risk']:
        with pytest.raises(core.InvalidQueryError):
            core.get_draws(cause_list, [measure], locations)


def test_get_relative_risk(mocker):
    gbd_mock = mocker.patch("vivarium_inputs.core.gbd")
    gbd_mock.get_age_group_id.return_value = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 30, 31, 32, 235]
    draw_cols = [f"rr_{i}" for i in range(10)]
    rr_maps = {'year_id': [1990, 1995, 2000], 'location_id': [1], 'sex_id': [1, 2], 'age_group_id': [4, 5],
               'risk_id': [240], 'cause_id': [302], 'parameter': ['cat1', 'cat2', 'cat3', 'cat4'],
               'morbidity': [1], 'mortality': [1], 'rei_id ': [240], 'modelable_entity_id': [9082], 'metric_id': [3]}

    rr_ = pd.DataFrame(columns=draw_cols, index=pd.MultiIndex.from_product([*rr_maps.values()], names=[*rr_maps.keys()]))
    rr_[draw_cols] = np.random.random_sample((len(rr_), 10)) * 10
    gbd_mock.get_relative_risks.return_value = rr_.reset_index()
    get_rr = core._get_relative_risk([risk_factors.child_wasting], [1])
    whole_age_groups = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 30, 31, 32, 235]
    missing_age_groups = list(set(whole_age_groups) - set(rr_maps['age_group_id']))
    missing_rr_maps = rr_maps.copy()
    missing_rr_maps['age_group_id'] = missing_age_groups

    del missing_rr_maps['morbidity']
    del missing_rr_maps['mortality']
    del missing_rr_maps['metric_id']
    del missing_rr_maps['modelable_entity_id']

    missing_rr = pd.DataFrame(1.0, columns=[f"draw_{i}" for i in range(10)],
                              index=pd.MultiIndex.from_product([*missing_rr_maps.values()],
                              names=[*missing_rr_maps.keys()]))
    rr_ = rr_.rename(columns={f'rr_{i}': f'draw_{i}' for i in range(10)})
    rr_ = rr_.reset_index(['morbidity', 'mortality', 'metric_id', 'modelable_entity_id'])
    rr_ = rr_.drop(['morbidity', 'mortality', 'metric_id', 'modelable_entity_id'], axis=1)
    expected_rr = rr_.append(missing_rr).sort_index().reset_index()
    get_rr = get_rr[expected_rr.columns]

    pd.util.testing.assert_frame_equal(expected_rr, get_rr)


def test_get_population_attributable_fraction(mocker):
    id_mock = mocker.patch("vivarium_inputs.core._get_ids_for_measure")
    gbd_mock = mocker.patch("vivarium_inputs.core.gbd")
    gbd_mock.get_age_group_id.return_value = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 30, 31, 32, 235]
    draw_cols = [f"draw_{i}" for i in range(10)]
    paf_maps = {'year_id': [1990, 1995, 2000], 'location_id': [1], 'sex_id': [1, 2], 'age_group_id': [4, 5],
                'rei_id': [84, 339, 238, 136, 240], 'cause_id': [302], 'measure_id':[3]}

    paf_ = pd.DataFrame(columns=draw_cols, index=pd.MultiIndex.from_product([*paf_maps.values()], names=[*paf_maps.keys()]))
    paf_[draw_cols] = np.random.random_sample((len(paf_), 10))
    gbd_mock.get_pafs.return_value = paf_.reset_index()

    get_paf = core._get_population_attributable_fraction([causes.diarrheal_diseases], [1])

    whole_age_groups = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 30, 31, 32, 235]
    missing_age_groups = list(set(whole_age_groups) - set(paf_maps['age_group_id']))
    missing_paf_maps = paf_maps.copy()
    missing_paf_maps['age_group_id'] = missing_age_groups
    del missing_paf_maps['measure_id']

    missing_paf = pd.DataFrame(0.0, columns=draw_cols,
                               index=pd.MultiIndex.from_product([*missing_paf_maps.values()],
                                                                names=[*missing_paf_maps.keys()]))
    paf_ = paf_.reset_index('measure_id')
    paf_ = paf_.drop('measure_id', axis=1)
    expected_paf = paf_.append(missing_paf).sort_index().reset_index()
    expected_paf = expected_paf.rename(columns={'rei_id':'risk_id'})
    get_paf = get_paf[expected_paf.columns]

    pd.util.testing.assert_frame_equal(expected_paf, get_paf)



import pytest
from hypothesis import given, assume
import hypothesis.strategies as st
from hypothesis.extra.pandas import data_frames, column

from gbd_mapping import Cause, Risk, Etiology, Sequela

from vivarium_inputs.utilities import get_age_group_midpoint_from_age_group_id
from vivarium_inputs.data_artifact import normalize

CAUSE_MEASURES = ["death", "prevalence", "incidence", "population_attributable_fraction",
                  "cause_specific_mortality", "excess_mortality", "remission"]
RISK_MEASURES = ["exposure", "exposure_standard_deviation", "relative_risk"]
POPULATION_ENTITY_PATHS = ["population.structure", "population.age_bins",
                           "population.theoretical_minimum_risk_life_expectancy"]


@st.composite
def measure(draw, entities, measures):
    measure = draw(st.sampled_from(measures))
    entity = draw(st.sampled_from(list(entities)))
    assume(entity is not None)
    entity_type = {
            Cause: "cause",
            Risk: "risk_factor",
            Etiology: "etiology",
            Sequela: "sequela",
    }[type(entity)]
    return f"{entity_type}.{entity.name}.{measure}"


@st.composite
def measure_dataframes(draw):
    columns = draw(st.sets(st.sampled_from([
                                 ("location", str),
                                 ("measure", str),
                                 ("year_id", int),
                                 ("sex_id", int),
                                 ("age_group_id", int)]
    )))
    columns = [column(c[0], dtype=c[1]) for c in columns] + [column(f"draw_{i}", dtype=float) for i in range(10)]
    df = draw(data_frames(columns=columns))
    assume(not df.empty)
    return df


@pytest.mark.skip("Cluster")
@given(measure_dataframes())
def test__normalize(data):
    normed = normalize(data)

    original_draw_count = len([c for c in data.columns if c.startswith("draw_")])
    normalized_draw_count = len(normed.draw.unique())
    assert original_draw_count == normalized_draw_count

    if "age_group_ids" in data:
        assert "age" in normed
        assert "age_group_start" in normed
        assert "age_group_end" in normed
        no_draw_orig = data[[c for c in data.columns if not c.startswith("draw_")]].unique()
        no_draw_norm = normed[[c for c in normed.columns if c != "draw"]].unique()
        assert no_draw_norm == get_age_group_midpoint_from_age_group_id(no_draw_orig)




from typing import Union

from gbd_mapping.id import meid
from gbd_mapping.base_template import GbdRecord, ModelableEntity


class HealthcareEntity(ModelableEntity):
    """Container for healthcare system GBD ids and data."""
    __slots__ = ('name', 'kind', 'gbd_id', 'utilization', 'cost')

    def __init__(self,
                 name: str,
                 kind: str,
                 gbd_id: Union[meid, None],
                 utilization: meid=None,
                 ):
        super().__init__(name=name, kind=kind, gbd_id=gbd_id)
        self.utilization = utilization


class HealthcareEntities(GbdRecord):
    """Holder of healthcare modelable entities"""
    __slots__ = ('outpatient_visits', 'inpatient_visits')

    def __init__(self,
                 outpatient_visits: HealthcareEntity,
                 inpatient_visits: HealthcareEntity):

        super().__init__()
        self.outpatient_visits = outpatient_visits
        self.inpatient_visits = inpatient_visits


class HealthTechnology(ModelableEntity):
    """Container for health technology ids and data."""
    __slots__ = ('name', 'kind', 'gbd_id')


class HealthTechnologies(GbdRecord):
    """Holder of healthcare technologies."""
    __slots__ = ('hypertension_drugs', )

    def __init__(self,
                 hypertension_drugs: HealthTechnology):
        self.hypertension_drugs = hypertension_drugs


healthcare_entities = HealthcareEntities(
    outpatient_visits=HealthcareEntity(
        name='outpatient_visits',
        kind='healthcare_entity',
        gbd_id=meid(10333),
        utilization=meid(10333),
    ),
    inpatient_visits=HealthcareEntity(
        name='inpatient_visits',
        kind='healthcare_entity',
        gbd_id=meid(10334),
        utilization=meid(10334),
    )
)

health_technologies = HealthTechnologies(
    hypertension_drugs=HealthTechnology(
        name='hypertension_drugs',
        kind='health_technology',
        gbd_id=None,
    )
)



from bdb import BdbQuit
import logging
import getpass
import pathlib
import argparse
import subprocess

import click

from vivarium.framework.configuration import build_model_specification
from vivarium.framework.plugins import PluginManager
from vivarium.interface.interactive import InteractiveContext
from vivarium.config_tree import ConfigTree


@click.command()
@click.argument('model_specification', type=click.Path(dir_okay=False,
                readable=True))
@click.argument('locations', nargs=-1)
@click.option('--project', '-P', default='proj_cost_effect',
              help='Cluster project under which the job will '
                   'be submitted. Defaults to proj_cost_effect')
@click.option('--output-root', '-o', type=click.Path(file_okay=False, writable=True),
              help="Directory to save artifact to. "
                   "Overwrites model specification file")
@click.option('--append', '-a', is_flag=True,
              help="Preserve existing artifact and append to it")
@click.option('--verbose', '-v', is_flag=True,
              help="Turn on debug mode for logging")
def build_artifact(model_specification, locations, project,
                   output_root, append, verbose):
    """
    build_artifact is a program for building data artifacts from a
    MODEL_SPECIFICATION file. The work is offloaded to the cluster
    under the "proj_cost_effect" project unless a different project is
    specified. Multiple, optional LOCATIONS can be provided to overwrite
    the configuration file. For locations containing spaces, replace the
    space with an underscore and surround any locations containing 
    apostrophes with double quotes, e.g.:

    build_artifact examply.yaml Virginia Pennsylvania New_York "Cote_d'Ivoire"

    Any artifact.path specified in the configuration file is guaranteed to
    be overwritten either by the optional output_root or a predetermined path
    based on user: /ihme/scratch/users/{user}/vivarium_artifacts

    This script necessarily offloads work to the cluster, and so requires being
    run in the cluster environment. To run locally, execute this script directly,
    as in `python cli.py`. The API is the same, and help can be accessed using
    -h / --help.
    """

    config_path = pathlib.Path(model_specification).resolve()
    python_context_path = pathlib.Path(subprocess.getoutput("which python")).resolve()
    script_path = pathlib.Path(__file__).resolve()

    script_args = f"{script_path} {config_path} "
    if output_root:
        script_args += f"--output_root {output_root} "
    if append:
        script_args += f"--append "
    if verbose:
        script_args += f"--verbose "

    num_locations = len(locations)
    if num_locations > 0:
        script_args += "--location {}"
        for i, location in enumerate(locations):
            location = location.replace("'", "-")
            job_name = f"{config_path.stem}_{location}_build_artifact"
            command = build_submit_command(python_context_path, job_name,
                                           project,
                                           script_args.format(location))
            click.echo(f"submitting job {i+1} of {num_locations} ({job_name})")
            submit_job(command, job_name)
    else:
        job_name = f"{config_path.stem}_build_artifact"
        command = build_submit_command(python_context_path, job_name,
                                       project, script_args)
        click.echo(f"submitting job {job_name}")
        submit_job(command, job_name)


def submit_job(command: str, name: str):
    """Submit a qsub command to the shell and report the result.

    Parameters
    ----------
    command
        A string containing a qsub job command
    name
        The name of the job described by `command`

    Returns
    -------
        None
    """

    exitcode, response = subprocess.getstatusoutput(command)
    if exitcode:
        click.secho(f"submission of {name} failed with exit code {exitcode}: {response}",
                    fg='red')
    else:
        click.secho(f"submission of {name} succeeded: {response}", fg='green')


def build_submit_command(python_context_path: str, job_name: str, project: str,
                         script_args: str, slots: int = 2) -> str:
    """Construct a valid qsub job command string.

    Parameters
    ----------
    python_context_path
        The full path to a python executable under which the job will be
        executed
    job_name
        The name of the job
    project
        The cluster project to run the job under
    script_args
        A string comprised of the full path to the script to be executed
        followed by its arguments
    slots
        The number of slots with which to execute the job

    Returns
    -------
    str
        A valid qsub command string
    """

    return (f"qsub -N {job_name} -P {project} -pe multi_slot {slots} " +
            f"-b y {python_context_path} " + script_args)


def _build_artifact():
    """Build a data artifact from a configuration file and optional arguments.

    This function parses command line input for arguments and builds a data
    artifact accordingly. It is meant for use by the click executable
    build_artifact, not as an outward-facing interface. Command line arguments
    are documented and can be viewed by calling this file with the -h flag.

    Returns
    -------
        None
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('model_specification', type=str,
                        help="path to a model_specification file")
    parser.add_argument('--output-root', '-o', type=str, required=False,
                        help="directory to save artifact to. "
                             "Overwrites model_specification file")
    parser.add_argument('--location', type=str, required=False,
                        help="location to get data for. "
                             "Overwrites model_specification file")
    parser.add_argument('--append', '-a', action="store_true",
                        help="Preserve existing artifact and append to it")
    parser.add_argument('--verbose', '-v', action='store_true',
                        help="Turn on debug mode for logging")
    parser.add_argument('--pdb', action='store_true')
    args = parser.parse_args()

    _setup_logging(args.output_root, args.verbose, args.location,
                   args.model_specification, args.append)

    try:
        main(args.model_specification, args.output_root, args.location, args.append)
    except (BdbQuit, KeyboardInterrupt):
        raise
    except Exception as e:
        logging.exception("Uncaught exception: %s", e)
        if args.pdb:
            import pdb
            import traceback
            traceback.print_exc()
            pdb.post_mortem()
        else:
            raise


def main(model_specification_file, output_root, location, append):
    model_specification = build_model_specification(model_specification_file)
    model_specification.plugins.optional.update({
        "data": {
            "controller": "vivarium_inputs.data_artifact.ArtifactBuilder",
            "builder_interface": "vivarium_public_health.dataset_manager.ArtifactManagerInterface",
        }})

    logging.debug("Configuring simulation")
    plugin_config = model_specification.plugins
    component_config = model_specification.components
    simulation_config = model_specification.configuration

    simulation_config.input_data.location = get_location(location, simulation_config)
    simulation_config.input_data.artifact_path = get_output_path(model_specification_file,
                                                                 output_root, location)
    simulation_config.input_data.append_to_artifact = append

    plugin_manager = PluginManager(plugin_config)
    component_config_parser = plugin_manager.get_plugin('component_configuration_parser')
    components = component_config_parser.get_components(component_config)
    
    logging.debug("Setting up simulation")
    simulation = InteractiveContext(simulation_config, components, plugin_manager)
    simulation.setup()


def get_location(location_arg: str, configuration: ConfigTree) -> str:
    """Resolve the model location

    This function takes in to account the model configuration and the passed
    location argument. User-passed arguments supercede the configuration.

    Parameters
    ----------
    location_arg
        the location argument passed to the click executable
    configuration
        A model configuration object

    Returns
    -------
    str
        The resolved location name
    """

    if location_arg:
        return location_arg.replace('_', ' ').replace("-", "'")
    elif contains_location(configuration):
        return configuration.input_data.location
    else:
        raise argparse.ArgumentError("specify a location or include "
                                     "configuration.input_data.location "
                                     "in model specification")


def contains_location(configuration: ConfigTree) -> bool:
    """Check if location is specified in the configuration

    Parameters
    ----------
    configuration
        A model configuration

    Returns
    -------
    bool
        True if a location is specified else False
    """

    return ('input_data' in configuration and
            'location' in configuration.input_data and
            configuration.input_data.location)


def get_output_path(configuration_arg: str, output_root_arg: str,
                    location_arg: str) -> str:
    """Resolve the correct model output path

    Takes in to account the model specification and passed arguments.
    User-passed arguments supercede the configuration. Guaranteed to
    clobber the configuration output path with either what the user
    provided or a default. The file name is taken from the model
    configuration file name. Location info is added if a location argument
    was provided.

    Parameters
    ----------
    configuration_arg
        Path to the model configuration file
    output_root_arg
        The output_root argument passed to the click executable
    location_arg
        The location argument passed to the click executable

    Returns
    -------
        A PathLike object containing the path to the output hdf file
    """

    configuration_path = pathlib.Path(configuration_arg)

    output_base = get_output_base(output_root_arg)

    if location_arg:
        output_path = output_base / (configuration_path.stem + f'_{location_arg}.hdf')
    else:
        output_path = output_base / (configuration_path.stem + '.hdf')
    return str(output_path)


def get_output_base(output_root_arg: str) -> pathlib.Path:
    """Resolve the correct output directory

    Defaults to /ihme/scratch/users/{user}/vivarium_artifacts/
    if no user passed output directory. Makes output directory
    if doesn't already exist.

    Parameters
    ----------
    output_root_arg
        The output_root argument passed to the click executable

    Returns
    -------
        A PathLike object containing the path to the output directory
    """

    if output_root_arg:
        output_base = pathlib.Path(output_root_arg).resolve()
    else:
        output_base = (pathlib.Path('/share') / 'scratch' / 'users' /
                       getpass.getuser() / 'vivarium_artifacts')

    if not output_base.is_dir():
        output_base.mkdir(parents=True)

    return output_base


def _setup_logging(output_root, verbose, location,
                   model_specification, append):
    """ Setup logging to write to a file in the output directory

    Log file named as {model_specification}_{location}_build_artifact.log
    to match naming format of qsubbed jobs. File saved in output directory
    (either passed by user or default)/logs. Raises error if that output
    directory is not found.

    """

    # this will raise an error if the passed output_root doesn't exist which gets
    # printed to console b/c logging isn't set up yet
    output_log_dir = get_output_base(output_root) / 'logs'

    if not output_log_dir.is_dir():
        output_log_dir.mkdir()

    log_level = logging.DEBUG if verbose else logging.ERROR
    log_tag = f'_{location}' if location is not None else ''
    spec_name = pathlib.Path(model_specification).resolve().stem
    log_name = f'{output_log_dir}/{spec_name}{log_tag}_build_artifact.log'

    logging.basicConfig(level=log_level,
                        format='%(asctime)s %(levelname)-8s %(message)s',
                        datefmt="%m-%d-%y %H:%M",
                        filename=log_name,
                        filemode='a' if append else 'w')


if __name__ == "__main__":
    _build_artifact()



"""Module containing functions that standardize the format of GBD outputs."""
from typing import Mapping

import pandas as pd

from vivarium_inputs.core import UnhandledDataError


def standardize_dimensions(data: pd.DataFrame, dimensions: pd.MultiIndex,
                           fill_na_value: Mapping[str, float]=None) -> pd.DataFrame:
    """Take input data and make it dense over the specified dimensions.

    The behavior of the function depends on which dimension is being considered and the nature of the sparsity:

    Cases:
    1) dimension missing - nothing needed
    2) fully dense in the expected range - nothing needed
    3) some sparsity:
       a) in the year dimension: interpolate
       b) in age/sex dimension and dense for a contiguous subset of the expected range and missing elsewhere: fill
    4) all others - panic

    Parameters
    ----------
    data :
        The data to standardize
    dimensions :
        A multi index whose individual index columns must be some subset of {'age', 'sex', 'year'}
        which represent dimensions of the data; and whose levels provide the expected extent of
        the data in those dimensions.
    fill_na_value :
        A mapping between

    Returns
    -------
        A dataframe with either full extent or no extent in all of the given dimensions.

    Raises
    ------
    UnhandledDataError :
        If the data is somehow malformed in a manner we don't deal with.
    """
    dimensions = dimensions.to_frame().reset_index(drop=True)

    dimensions = pd.concat([dimensions.assign(measure=measure) for measure in data.measure.unique()], ignore_index=True)
    draw_columns = [c for c in data.columns if 'draw_' in c]

    assert set(dimensions.columns) <= {'age_group_id', 'sex', 'year', 'measure', 'location'}
    assert set(data.columns) <= set(dimensions.columns.tolist() + draw_columns)
    assert set(data.measure.unique()) <= fill_na_value.keys()

    # Case 1: Remove any dimensions not present in the data.
    applicable_dimensions = dimensions[[c for c in dimensions.columns if c in data]].copy()
    # Case 4: Check preconditions and panic if it's an unhandled case.
    verify_well_formed(data, applicable_dimensions)

    # Case 3a: Interpolate over year.
    if 'year' in applicable_dimensions.columns:
        data = interpolate_years(data, applicable_dimensions)

    # Case 3b: Fill in missing data
    dimension_columns, extents = zip(*applicable_dimensions.items())
    expected_index = pd.MultiIndex.from_arrays(extents, names=dimension_columns)
    data = data.set_index(list(dimension_columns))

    missing = expected_index.difference(data.index)
    if not missing.empty:
        to_add = pd.DataFrame(columns=draw_columns, index=missing)
        to_add = to_add.reset_index().set_index(list(set(dimension_columns) - {'measure'})).sort_index()
        data = data.reset_index().set_index(list(set(dimension_columns) - {'measure'})).sort_index()
        for measure, fill in fill_na_value.items():
            to_add.loc[to_add['measure'] == measure, draw_columns] = fill
        data = data.append(to_add)

    return data.reset_index()


def verify_well_formed(data: pd.DataFrame, dimensions: pd.DataFrame):
    for dimension in dimensions.columns:
        if dimension == 'sex':
            continue
        existing = data[dimension].sort_values().drop_duplicates()
        expected = dimensions[dimension].sort_values().drop_duplicates()

        contiguous_overlap = ((expected >= existing.min()) & (expected <= existing.max()))
        if dimension == 'age_group_id' and not set(existing) == set(expected[contiguous_overlap]):
            raise UnhandledDataError(f'The data is malformed in the {dimension} dimension.')

        if dimension == 'year' and (existing.min() > expected.min() or existing.max() < expected.max()):
            raise UnhandledDataError("Cannot interpolate years without data on both edges")


def interpolate_years(data: pd.DataFrame, dimensions: pd.DataFrame) -> pd.DataFrame:

    existing_extent = data['year'].sort_values().drop_duplicates()
    expected_extent = dimensions['year'].sort_values().drop_duplicates()

    out = []
    for year in expected_extent:
        if year in existing_extent.values:
            out.append(data[data['year'] == year])
        else:
            from core_maths.interpolate import interpolate
            index_columns = list(dimensions.columns)
            value_columns = list(data.columns.difference(index_columns))

            previous_year = int(existing_extent[existing_extent < year].iloc[-1])
            next_year = int(existing_extent[existing_extent > year].iloc[0])
            start = data[data['year'] == previous_year].sort_values(index_columns).reset_index(drop=True)
            end = data[data['year'] == next_year].sort_values(index_columns).reset_index(drop=True)

            interpolated = interpolate(start, end, index_columns, 'year', value_columns, previous_year, next_year)
            interpolated = interpolated[interpolated['year'] == year]
            out.append(interpolated)

    return pd.concat(out, ignore_index=True)


def select_draw_data(data, draw, column_name, src_column=None):
    if column_name:
        if src_column is not None:
            if isinstance(src_column, str):
                column_map = {src_column.format(draw=draw): column_name}
            else:
                column_map = {src.format(draw=draw): dest for src, dest in zip(src_column, column_name)}
        else:
            column_map = {'draw_{draw}'.format(draw=draw): column_name}

        # if 'measure' is in columns, then keep it, else do
        # not keep it (need measure for the relative risk estimations)
        if 'parameter' in data.columns:
            keep_columns = ['year_id', 'age', 'sex_id', 'parameter'] + list(column_map.keys())
        else:
            keep_columns = ['year_id', 'age', 'sex_id'] + list(column_map.keys())

        data = data[keep_columns]
        data = data.rename(columns=column_map)

        return normalize_for_simulation(data)
    return data


def normalize_for_simulation(df):
    """
    Parameters
    ----------
    df : DataFrame
        dataframe to change

    Returns
    -------
    Returns a df with column year_id changed to year, and year_start and year_end
    created as bin ends around year_id with year_start set to year_id;
    sex_id changed to sex, and sex values changed from 1 and 2 to Male and Female

    Notes
    -----
    Used by -- load_data_from_cache

    Assumptions -- None

    Questions -- None

    Unit test in place? -- Yes
    """
    if "sex_id" in df:
        if set(df["sex_id"]) == {3}:
            df_m = df.copy()
            df_f = df.copy()
            df_m['sex'] = 'Male'
            df_f['sex'] = 'Female'
            df = pd.concat([df_m, df_f], ignore_index=True)
        else:
            df["sex"] = df.sex_id.map({1: "Male", 2: "Female", 3: "Both"}).astype(
                pd.api.types.CategoricalDtype(categories=["Male", "Female", "Both"], ordered=False))

        df = df.drop("sex_id", axis=1)

    if "year_id" in df:
        # FIXME: use central comp interpolation tools
        if 2006 in df.year_id.unique() and 2007 not in df.year_id.unique():
            df = df.loc[(df.year_id != 2006)]

        df = df.rename(columns={"year_id": "year"})
        idx = df.index

        mapping = df[['year']].drop_duplicates().sort_values(by="year")
        mapping['year_start'] = mapping['year']
        mapping['year_end'] = mapping['year'].shift(-1).fillna(mapping.year.max()+1)

        df = df.set_index("year", drop=False)
        mapping = mapping.set_index("year", drop=False)

        df[["year_start", "year_end"]] = mapping[["year_start", "year_end"]]

        df = df.set_index(idx)

    return df


def get_age_group_midpoint_from_age_group_id(df):
    """Creates an "age" column from the "age_group_id" column

    Parameters
    ----------
    df: df for which you want an age column that has an age_group_id column

    Returns
    -------
    df with an age column

    Notes
    -----
    Assumptions -- We assume that using a midpoint of age 82.5 for the 80+ year old age group is
    ok for the purposes of Vivarium. Everett proposed that we could get the life expectancy at age 80
    for each location and use that as the midpoint for the 80+ group, but Abie suggested that we
    keep the midpoint as 82.5 for now. GBD populations have data for each age group up until the
    age 95+ age group, at which point I'm assuming we can use 97.5 as the midpoint.
    """
    import vivarium_gbd_access.gbd as gbd
    if df.empty:
        df['age'] = 0
        return df

    df = df.copy()
    idx = df.index
    mapping = gbd.get_age_bins()
    mapping = mapping.set_index('age_group_id')
    mapping['age'] = mapping[['age_group_years_start', 'age_group_years_end']].mean(axis=1)

    df = df.set_index('age_group_id')
    df[['age', 'age_group_start', 'age_group_end']] = mapping[['age', 'age_group_years_start', 'age_group_years_end']]

    # Assumption: We're using 82.5 as the midpoint for the age 80+ age group. May want to change in the future.
    df.loc[df.age == 102.5, 'age'] = 82.5

    df = df.set_index(idx)

    return df



import pandas as pd

from vivarium_inputs.utilities import normalize_for_simulation, get_age_group_midpoint_from_age_group_id
from vivarium_inputs import core


def normalize(data: pd.DataFrame) -> pd.DataFrame:
    """Remove GBD specific column names and concepts and make the dataframe long over draws."""
    assert not data.empty
    data = normalize_for_simulation(data)
    if "age_group_id" in data:
        data = get_age_group_midpoint_from_age_group_id(data)
    draw_columns = [c for c in data.columns if "draw_" in c]
    index_columns = [c for c in data.columns if "draw_" not in c]
    data = pd.melt(data, id_vars=index_columns, value_vars=draw_columns, var_name="draw")
    data["draw"] = data.draw.str.partition("_")[2].astype(int)

    estimation_years = core.get_estimation_years()
    year_start, year_end = min(estimation_years), max(estimation_years)
    if "year" in data:
        data = data.loc[(data.year >= year_start) & (data.year <= year_end)]
    return data



#!/usr/bin/env python
import os

from setuptools import setup, find_packages

if __name__ == "__main__":
    base_dir = os.path.dirname(__file__)
    src_dir = os.path.join(base_dir, "src")

    about = {}
    with open(os.path.join(src_dir, "vivarium_inputs", "__about__.py")) as f:
        exec(f.read(), about)

    with open(os.path.join(base_dir, "README.rst")) as f:
        long_description = f.read()

    install_requirements = [
        'numpy',
        'scipy',
        'pandas',
        'click',
        'joblib',
        'tables',
        'vivarium',
        'vivarium_public_health',
        'gbd_mapping',
    ]

    data_requires = [
        'vivarium-gbd-access>=1.0.6',
    ]

    test_requirements = [
        'pytest',
        'pytest-mock',
        'hypothesis',
    ]

    doc_requirements = [
        'sphinx',
        'sphinx-autodoc-typehints',
        'sphinx-rtd-theme',
    ]

    setup(
        name=about['__title__'],
        version=about['__version__'],

        description=about['__summary__'],
        long_description=long_description,
        url=about["__uri__"],

        author=about["__author__"],
        author_email=about["__email__"],

        package_dir={'': 'src'},
        packages=find_packages(where='src'),
        include_package_data=True,

        install_requires=install_requirements,
        tests_require=test_requirements,
        extras_require={
            'docs': doc_requirements,
            'test': test_requirements,
            'data': data_requires,
            'dev': doc_requirements + test_requirements + data_requires,
        },

        entry_points='''
            [console_scripts]
            build_artifact=vivarium_inputs.data_artifact.cli:build_artifact
        ''',

        zip_safe=False,
    )



from vivarium.config_tree import ConfigTree  # Just for typing info.
from gbd_mapping import covariates

try:
    from vivarium_gbd_access.utilities import get_input_config
except ModuleNotFoundError:
    from vivarium.framework.configuration import build_simulation_configuration

    def get_input_config(override_config):
        config = build_simulation_configuration()
        config.update(override_config)
        config.update({
            'input_data': {
                'location': 'Kenya',
                'input_draw_number': 0,
            }
        })

        return config


from vivarium_inputs import core
from vivarium_inputs.utilities import (select_draw_data, get_age_group_midpoint_from_age_group_id,
                                       normalize_for_simulation)
from vivarium_inputs.mapping_extension import healthcare_entities

__all__ = [
    "get_prevalence", "get_incidence", "get_remission", "get_cause_specific_mortality", "get_excess_mortality",
    "get_disability_weight", "get_relative_risk", "get_exposure", "get_exposure_standard_deviation",
    "get_population_attributable_fraction", "get_ensemble_weights", "get_populations", "get_age_bins",
    "get_theoretical_minimum_risk_life_expectancy", "get_subregions", "get_outpatient_visit_costs",
    "get_inpatient_visit_costs", "get_inpatient_visit_costs",
    "get_age_specific_fertility_rates", "get_live_births_by_sex", "get_dtp3_coverage",
    "get_protection", "get_healthcare_annual_visits",
]


def _clean_and_filter_data(data, draw_number, column_name):
    key_cols = [c for c in data.columns if 'draw' not in c]
    data = data[key_cols + [f'draw_{draw_number}']]
    data = get_age_group_midpoint_from_age_group_id(data)
    return select_draw_data(data, draw_number, column_name)


####################################
# Measures for cause like entities #
####################################

def get_prevalence(entity, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([entity], ['prevalence'], [config.input_data.location]).drop('measure', 'columns')
    return _clean_and_filter_data(data, config.input_data.input_draw_number, 'prevalence')


def get_incidence(entity, override_config: ConfigTree=None):
    config = get_input_config(override_config)
    data = core.get_draws([entity], ['incidence'], [config.input_data.location]).drop('measure', 'columns')
    return _clean_and_filter_data(data, config.input_data.input_draw_number, 'rate')


def get_remission(cause, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([cause], ['remission'], [config.input_data.location]).drop('measure', 'columns')
    return _clean_and_filter_data(data, config.input_data.input_draw_number, 'rate')


def get_cause_specific_mortality(cause, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([cause], ['cause_specific_mortality'],
                          [config.input_data.location]).drop('measure', 'columns')
    return _clean_and_filter_data(data, config.input_data.input_draw_number, 'rate')


def get_excess_mortality(cause, override_config: ConfigTree=None):
    config = get_input_config(override_config)
    data = core.get_draws([cause], ['excess_mortality'], [config.input_data.location]).drop('measure', 'columns')
    return _clean_and_filter_data(data, config.input_data.input_draw_number, 'rate')


def get_disability_weight(sequela, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([sequela], ['disability_weight'], [config.input_data.location]).drop('measure', 'columns')
    return float(data[f'draw_{config.input_data.input_draw_number}'])


####################################
# Measures for risk like entities  #
####################################


def get_relative_risk(entity, cause, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([entity], ['relative_risk'], [config.input_data.location]).drop('measure', 'columns')
    data = data[data['cause_id'] == cause.gbd_id]
    return _clean_and_filter_data(data, config.input_data.input_draw_number, 'relative_risk')


def get_exposure(risk, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([risk], ['exposure'], [config.input_data.location]).drop('measure', 'columns')
    data = _clean_and_filter_data(data, config.input_data.input_draw_number, 'mean')
    # FIXME: This is here because FPG puts zeros in its unmodelled age groups unlike most other gbd risks
    data = data[data['mean'] != 0]
    return data


def get_exposure_standard_deviation(risk, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([risk], ['exposure_standard_deviation'],
                          [config.input_data.location]).drop('measure', 'columns')
    data = _clean_and_filter_data(data, config.input_data.input_draw_number, 'standard_deviation')
    # FIXME: This is here because FPG puts zeros in its unmodelled age groups unlike most other gbd risks
    data = data[data['standard_deviation'] != 0]
    return data


def get_population_attributable_fraction(entity, risk, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([entity], ['population_attributable_fraction'],
                          [config.input_data.location]).drop('measure', 'columns')
    data = data[data['risk_id'] == risk.gbd_id]
    return _clean_and_filter_data(data, config.input_data.input_draw_number, 'population_attributable_fraction')


def get_ensemble_weights(risk, override_config=None):
    return core.get_ensemble_weights([risk])


#######################
# Other kinds of data #
#######################


def get_populations(override_config=None, location=None):
    config = get_input_config(override_config)
    if location:
        data = core.get_populations([location])
    else:
        data = core.get_populations([config.input_data.location])
    data = get_age_group_midpoint_from_age_group_id(data)
    data = normalize_for_simulation(data)
    return data


def get_age_bins():
    return core.get_age_bins()


def get_theoretical_minimum_risk_life_expectancy():
    return core.get_theoretical_minimum_risk_life_expectancy()


def get_subregions(override_config=None):
    config = get_input_config(override_config)
    return core.get_subregions([config.input_data.location])


def get_outpatient_visit_costs(override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([healthcare_entities.outpatient_visits],
                          ['cost'], [config.input_data.location]).drop('measure', 'columns')
    data = data[['year_id', f'draw_{config.input_data.input_draw_number}']]
    return data.rename(columns={'year_id':'year', f'draw_{config.input_data.input_draw_number}': 'cost'})


def get_inpatient_visit_costs(override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([healthcare_entities.inpatient_visits],
                          ['cost'], [config.input_data.location]).drop('measure', 'columns')
    data = data[['year_id', f'draw_{config.input_data.input_draw_number}']]
    return data.rename(columns={'year_id':'year', f'draw_{config.input_data.input_draw_number}': 'cost'})


def get_age_specific_fertility_rates(override_config=None):
    config = get_input_config(override_config)
    data = core.get_covariate_estimates([covariates.age_specific_fertility_rate], [config.input_data.location])
    data = get_age_group_midpoint_from_age_group_id(data)
    data = normalize_for_simulation(data)
    return data.loc[data.sex == 'Female', ['age', 'year', 'mean_value']].rename(columns={'mean_value': 'rate'})


def get_live_births_by_sex(override_config=None):
    config = get_input_config(override_config)
    data = core.get_covariate_estimates([covariates.live_births_by_sex], [config.input_data.location])
    data = data[['sex_id', 'year_id', 'mean_value', 'lower_value', 'upper_value']]
    return normalize_for_simulation(data)


def get_dtp3_coverage(override_config=None):
    config = get_input_config(override_config)
    data = core.get_covariate_estimates([covariates.dtp3_coverage_proportion], [config.input_data.location])
    data = normalize_for_simulation(data)
    return data[['mean_value', 'lower_value', 'upper_value', 'year']]


def get_protection(treatment_technology, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([treatment_technology],
                          ['protection'], [config.input_data.location]).drop('measure', 'columns')
    data = data[['location', 'treatment_technology', f'draw_{config.input_data.input_draw_number}']]
    return data.rename(columns={f'draw_{config.input_data.input_draw_number}': 'protection'})


def get_healthcare_annual_visits(healthcare_entity, override_config=None):
    config = get_input_config(override_config)
    data = core.get_draws([healthcare_entity],
                          ['annual_visits'], [config.input_data.location]).drop('measure', 'columns')
    return _clean_and_filter_data(data, config.input_data.input_draw_number, 'annual_visits')



"""This module performs the core data transformations on GBD data and provides a basic API for data access."""
from typing import Iterable, Sequence, List, Union

import numpy as np
import pandas as pd

from gbd_mapping.id import scalar, UNKNOWN
from gbd_mapping.base_template import ModelableEntity
from gbd_mapping.cause import Cause, causes
from gbd_mapping.risk import Risk, risk_factors
from gbd_mapping.sequela import Sequela
from gbd_mapping.etiology import Etiology, etiologies
from gbd_mapping.coverage_gap import CoverageGap, coverage_gaps
from gbd_mapping.covariate import Covariate

try:
    import vivarium_gbd_access.gbd as gbd
except ModuleNotFoundError:
    class GbdDummy:
        def __getattr__(self, item):
            raise ModuleNotFoundError("Required package vivarium_gbd_access not found.")
    gbd = GbdDummy()


from vivarium_inputs.mapping_extension import HealthcareEntity, HealthTechnology


# Define GBD sex ids for usage with central comp tools.
MALE = [1]
FEMALE = [2]
COMBINED = [3]

name_measure_map = {'death': 1,
                    'DALY': 2,
                    'YLD': 3,
                    'YLL': 4,
                    'prevalence': 5,
                    'incidence': 6,
                    'remission': 7,
                    'excess_mortality': 9,
                    'proportion': 18,
                    'continuous': 19, }
gbd_round_id_map = {3: 'GBD_2015', 4: 'GBD_2016'}
age_restriction_map = {scalar(0.0): [2, None],
                       scalar(0.01): [3, 2],
                       scalar(0.10): [4, 3],
                       scalar(1.0): [5, 4],
                       scalar(5.0): [6, 5],
                       scalar(10.0): [7, 6],
                       scalar(15.0): [8, 7],
                       scalar(20.0): [9, 8],
                       scalar(30.0): [11, 10],
                       scalar(40.0): [13, 12],
                       scalar(45.0): [14, 13],
                       scalar(50.0): [15, 14],
                       scalar(55.0): [16, 15],
                       scalar(65.0): [18, 17],
                       scalar(95.0): [235, 32], }


class DataError(Exception):
    """Base exception for errors in data loading."""
    pass


class InvalidQueryError(DataError):
    """Exception raised when the user makes an invalid request for data (e.g. exposures for a sequela)."""
    pass


class UnhandledDataError(DataError):
    """Exception raised when we receive data from the databases that we don't know how to handle."""
    pass


class DataMissingError(DataError):
    """Exception raised when data has unhandled missing entries."""
    pass


class DuplicateDataError(DataError):
    """Exception raised when data has duplication in the index."""
    pass


def get_draws(entities: Sequence[ModelableEntity], measures: Iterable[str],
              locations: Iterable[str]) -> pd.DataFrame:
    """Gets draw level gbd data for each specified measure and entity.

    Parameters
    ----------
    entities:
        A list of data containers from the `gbd_mapping` package. The entities must all be the same
        type (e.g. all `gbd_mapping.Cause` objects or all `gbd_mapping.Risk` objects, etc.
    measures:
        A list of the GBD measures requested for the provided entities.
    locations:
        A list of locations to pull data for.

    Returns
    -------
    A table of draw level data for indexed by an entity, measure, and location combination as well as demographic data
    (age_group_id, sex_id, year_id) where appropriate.
    """
    measure_handlers = {
        'death': (_get_death, set()),
        'remission': (_get_remission, set()),
        'prevalence': (_get_prevalence, set()),
        'incidence': (_get_incidence, set()),
        'relative_risk': (_get_relative_risk, {'cause_id', 'parameter', }),
        'population_attributable_fraction': (_get_population_attributable_fraction, {'cause_id', }),
        'cause_specific_mortality': (_get_cause_specific_mortality, set()),
        'excess_mortality': (_get_excess_mortality, set()),
        'exposure': (_get_exposure, {'parameter', }),
        'exposure_standard_deviation': (_get_exposure_standard_deviation, {'risk_id', }),
        'annual_visits': (_get_annual_visits, {'modelable_entity_id', }),
        'disability_weight': (_get_disability_weight, set()),
        'cost': (_get_cost, set()),
    }

    location_ids = [get_location_ids_by_name()[name] for name in locations]

    data = []
    id_cols = set()
    for measure in measures:
        try:
            handler, id_columns = measure_handlers[measure]
        except KeyError:
            raise InvalidQueryError(f'No functions are available to pull data for measure {measure}')
        measure_data = handler(entities, location_ids)
        measure_data['measure'] = measure
        id_cols |= id_columns
        data.append(measure_data)
    data = pd.concat(data)

    id_cols |= _get_additional_id_columns(data, entities)
    key_columns = ['measure']
    for column in ['year_id', 'sex_id', 'age_group_id', 'location_id']:
        if column in data:
            key_columns.append(column)

    key_columns += list(id_cols)
    draw_columns = [f'draw_{i}' for i in range(0, 1000)]

    data = data[key_columns + draw_columns].reset_index(drop=True)
    _validate_data(data, key_columns)
    if "location_id" in data:
        data["location"] = data.location_id.apply(get_location_names_by_id().get)
        data = data.drop("location_id", "columns")

    return data


# TODO: Move to utilities.py
def _get_ids_for_measure(entities: Sequence[ModelableEntity], measure: str) -> List:
    """Selects the appropriate gbd id type for each entity and measure pair.

    Parameters
    ----------
    entities :
        A list of data containers from the `gbd_mapping` package. The entities must all be the same
        type (e.g. all `gbd_mapping.Cause` objects or all `gbd_mapping.Risk` objects, etc.
    measure :
        A list of the GBD measures requested for the provided entities.

    Returns
    -------
    A dictionary whose keys are the requested measures and whose values are sets of the appropriate
    GBD ids for use with central comp tools for the provided entities.

    Raises
    ------
    InvalidQueryError
        If the entities passed are inconsistent with the requested measures.
    """
    measure_types = {
        'death': (Cause, 'gbd_id'),
        'prevalence': ((Cause, Sequela), 'gbd_id'),
        'incidence': ((Cause, Sequela), 'gbd_id'),
        'exposure': ((Risk, CoverageGap), 'gbd_id'),
        'exposure_standard_deviation': ((Risk, CoverageGap), 'gbd_id'),
        'relative_risk': ((Risk, CoverageGap), 'gbd_id'),
        'population_attributable_fraction': ((Cause, Etiology, CoverageGap), 'gbd_id'),
        'cause_specific_mortality': ((Cause,), 'gbd_id'),
        'excess_mortality': ((Cause,), 'gbd_id'),
        'annual_visits': (HealthcareEntity, 'utilization'),
        'disability_weight': (Sequela, 'gbd_id'),
        'remission': (Cause, 'dismod_id'),
        'cost': ((HealthcareEntity, HealthTechnology), 'cost'),
    }

    if not all([isinstance(e, type(entities[0])) for e in entities]):
        raise InvalidQueryError("All entities must be of the same type")
    if measure not in measure_types.keys():
        raise InvalidQueryError(f"You've requested an invalid measure: {measure}")

    valid_types, id_attr = measure_types[measure]
    out = []
    for entity in entities:
        if isinstance(entity, valid_types):
            value = entity
            for sub_id in id_attr.split('.'):
                if value[sub_id] is not UNKNOWN:
                    value = value[sub_id]
                else:
                    raise InvalidQueryError(f"Entity {entity.name} has no data for measure '{measure}'")
            out.append(value)
        else:
            raise InvalidQueryError(f"Entity {entity.name} has no data for measure '{measure}'")

    return out


# TODO: Move to utilities.py
def _get_additional_id_columns(data, entities):
    id_column_map = {
        Cause: 'cause_id',
        Sequela: 'sequela_id',
        Covariate: 'covariate_id',
        Risk: 'risk_id',
        Etiology: 'etiology_id',
        CoverageGap: 'coverage_gap',
        HealthcareEntity: 'healthcare_entity',
        HealthTechnology: 'health_technology',
    }
    out = set()
    out.add(id_column_map[type(entities[0])])
    if isinstance(entities[0], CoverageGap) and data['measure'].all() in ['relative_risk']:
        out.add('risk_id')
    out |= set(data.columns) & set(id_column_map.values())
    return out


# TODO: Move to utilities.py
def _validate_data(data: pd.DataFrame, key_columns: Iterable[str]=None):
    """Validates that no data is missing and that the provided key columns make a valid (unique) index.

    Parameters
    ----------
    data:
        The data table to be validated.
    key_columns:
        An iterable of the column names used to uniquely identify a row in the data table.

    Raises
    ------
    DataMissingError
        If the data contains any null (NaN or NaT) values.
    DuplicatedDataError
        If the provided key columns are insufficient to uniquely identify a record in the data table.
    """

    #  check draw_cols only since we may have nan in coverage_gap data from aux_data
    draw_cols = [f'draw_{i}' for i in range(1000)]
    if np.any(data[draw_cols].isnull()):
        raise DataMissingError()

    if key_columns and np.any(data.duplicated(key_columns)):
        raise DuplicateDataError()

#####################
# get_draws helpers #
#####################
#
# These functions filter out erroneous measures and deal with special cases.
#

####################
# Cause-like stuff #
####################


def _get_death(entities, location_ids):
    measure_ids = _get_ids_for_measure(entities, 'death')
    death_data = gbd.get_codcorrect_draws(cause_ids=measure_ids, location_ids=location_ids)

    return death_data[death_data['measure_id'] == name_measure_map['death']]


def _get_remission(entities, location_ids):
    measure_ids = _get_ids_for_measure(entities, 'remission')
    remission_data = gbd.get_modelable_entity_draws(me_ids=measure_ids, location_ids=location_ids)

    id_map = {entity.dismod_id: entity.gbd_id for entity in entities}
    remission_data['cause_id'] = remission_data['modelable_entity_id'].replace(id_map)

    # FIXME: The sex filtering should happen in the reshaping step.
    correct_measure = remission_data['measure_id'] == name_measure_map['remission']
    correct_sex = remission_data['sex_id'] != COMBINED
    return remission_data[correct_measure & correct_sex]


def _get_prevalence(entities, location_ids):
    measure_ids = _get_ids_for_measure(entities, 'prevalence')
    entity_type = "cause" if isinstance(entities[0], Cause) else "sequela"
    measure_data = gbd.get_como_draws(entity_ids=measure_ids, location_ids=location_ids, entity_type=entity_type)

    # FIXME: The year filtering should happen in the reshaping step.
    correct_measure = measure_data['measure_id'] == name_measure_map['prevalence']
    correct_years = measure_data['year_id'].isin(gbd.get_estimation_years(gbd.GBD_ROUND_ID))
    return measure_data[correct_measure & correct_years]


def _get_incidence(entities, location_ids):
    measure_ids = _get_ids_for_measure(entities, 'incidence')
    entity_type = "cause" if isinstance(entities[0], Cause) else "sequela"
    measure_data = gbd.get_como_draws(entity_ids=measure_ids, location_ids=location_ids, entity_type=entity_type)

    # FIXME: The year filtering should happen in the reshaping step.
    correct_measure = measure_data['measure_id'] == name_measure_map['incidence']
    correct_years = measure_data['year_id'].isin(gbd.get_estimation_years(gbd.GBD_ROUND_ID))

    # get_draws returns "total incidence." We want hazard incidence.
    # scale by the number of people not sick to convert.

    incidence = measure_data[correct_measure & correct_years]
    prevalence = _get_prevalence(entities, location_ids)

    draw_columns = [col for col in incidence.columns if col.startswith('draw_')]

    incidence.set_index([f'{entity_type}_id', 'age_group_id', 'location_id', 'sex_id', 'year_id'], inplace=True)
    prevalence.set_index([f'{entity_type}_id', 'age_group_id', 'location_id', 'sex_id', 'year_id'], inplace=True)

    incidence[draw_columns] = incidence[draw_columns] / (1 - prevalence[draw_columns])

    return incidence.reset_index()


def _get_cause_specific_mortality(entities, location_ids):
    # FIXME: Mapping backword to name like this is awkward
    locations = [get_location_names_by_id()[location_id] for location_id in location_ids]
    deaths = get_draws(entities, ["death"], locations)
    deaths["location_id"] = deaths.location.apply(get_location_ids_by_name().get)
    deaths = deaths.drop("location", "columns")

    populations = get_populations(locations)
    populations = populations[populations['year_id'] >= deaths.year_id.min()]
    populations["location_id"] = populations.location.apply(get_location_ids_by_name().get)

    merge_columns = ['age_group_id', 'location_id', 'year_id', 'sex_id']
    key_columns = merge_columns + ['cause_id']
    draw_columns = [f'draw_{i}' for i in range(0, 1000)]

    df = deaths.merge(populations, on=merge_columns).set_index(key_columns)
    csmr = df[draw_columns].divide(df['population'], axis=0).reset_index()

    csmr = csmr[key_columns + draw_columns]
    _validate_data(csmr, key_columns)

    return _standardize_data(csmr, 0)


def _get_excess_mortality(entities, location_ids):
    # FIXME: Mapping backword to name like this is awkward
    locations = [get_location_names_by_id()[location_id] for location_id in location_ids]
    prevalences = get_draws(entities, ['prevalence'], locations).drop('measure', 'columns')
    csmrs = get_draws(entities, ['cause_specific_mortality'], locations).drop('measure', 'columns')

    key_columns = ['year_id', 'sex_id', 'age_group_id', 'location', 'cause_id']
    prevalences = prevalences.set_index(key_columns)
    csmrs = csmrs.set_index(key_columns)

    # In some cases CSMR is not zero for age groups where prevalence is, which leads to
    # crazy outputs. So enforce that constraint.
    # TODO: But is this the right place to do that?
    draw_columns = [f'draw_{i}' for i in range(1000)]
    csmrs[draw_columns] = csmrs[draw_columns].where(prevalences[draw_columns] != 0, 0)

    em = csmrs.divide(prevalences, axis='index').reset_index()
    em = em[em['sex_id'] != COMBINED]

    em["location_id"] = em.location.apply(get_location_ids_by_name().get)
    em = em.drop("location", "columns")

    return _standardize_data(em.dropna(), 0)


# TODO This should probably use the _get_ids_for_measure function but it doesn't quite fit
def _get_disability_weight(entities, _):
    if isinstance(entities[0], Sequela):
        disability_weights = gbd.get_auxiliary_data('disability_weight', 'sequela', 'all')
    else:
        raise InvalidQueryError("Only sequela have disability weights associated with them.")

    data = []
    for s in entities:

        if s.healthstate.gbd_id in disability_weights['healthstate_id'].values:
            df = disability_weights.loc[disability_weights.healthstate_id == s.healthstate.gbd_id, :]
        else:
            raise DataMissingError(f"No disability weight available for the sequela {s.name}")

        df['sequela_id'] = s.gbd_id
        data.append(df)

    data = pd.concat(data, ignore_index=True)

    return data.reset_index(drop=True)


###################
# Risk-like stuff #
###################

def _standardize_data(data: pd.DataFrame, fill_value: int) -> pd.DataFrame:
    # age_groups that we expect to exist for each risk
    whole_age_groups = gbd.get_age_group_id()

    sex_id = data.sex_id.unique()
    year_id = data.year_id.unique()
    location_id = data.location_id.unique()

    index_cols = ['year_id', 'location_id', 'sex_id', 'age_group_id']
    draw_cols = [c for c in data.columns if 'draw_' in c]

    other_cols = {c: data[c].unique() for c in data.dropna(axis=1).columns if c not in index_cols and c not in draw_cols}
    index_cols += [*other_cols.keys()]
    data = data.set_index(index_cols)

    # expected indexes to be in the data
    expected = pd.MultiIndex.from_product([year_id, location_id, sex_id, whole_age_groups] + [*other_cols.values()],
                                          names=(['year_id', 'location_id', 'sex_id', 'age_group_id'] + [
                                              *other_cols.keys()]))

    new_data = data.copy()
    missing = expected.difference(data.index)

    # assign dtype=float to prevent the artifact error with mixed dtypes
    to_add = pd.DataFrame({column: fill_value for column in draw_cols}, index=missing, dtype=float)

    new_data = new_data.append(to_add).sort_index()

    return new_data.reset_index()


def _standardize_all_age_groups(data: pd.DataFrame):
    if set(data.age_group_id) != {22}:
        return data
    whole_age_groups = gbd.get_age_group_id()
    missing_age_groups = set(whole_age_groups).difference(set(data.age_group_id))
    df = []
    for i in missing_age_groups:
        missing = data.copy()
        missing['age_group_id'] = i
        df.append(missing)

    return pd.concat(df, ignore_index=True)


def _get_relative_risk(entities: Iterable[Union[Risk, CoverageGap]], location_ids: Iterable[int]):

    # common pre-processing for entities from GBD
    def _pull_rr_data_from_gbd(measure_ids):
        data = gbd.get_relative_risks(risk_ids=measure_ids, location_ids=location_ids)
        data = data.rename(columns={f'rr_{i}': f'draw_{i}' for i in range(1000)})

        # FIXME: I'm passing because this is broken for zinc_deficiency, and I don't have time to investigate -J.C.
        # err_msg = ("Not all relative risk data has both the 'mortality' and 'morbidity' flag "
        #            + "set. This may not indicate an error but it is a case we don't explicitly handle. "
        #            + "If you need this risk, come talk to one of the programmers.")
        # assert np.all((measure_data.mortality == 1) & (measure_data.morbidity == 1)), err_msg

        data = data[data['morbidity'] == 1]  # FIXME: HACK
        del data['mortality']
        del data['morbidity']
        del data['metric_id']
        del data['modelable_entity_id']
        return data

    measure_ids = _get_ids_for_measure(entities, 'relative_risk')

    if isinstance(entities[0], Risk):
        if None in measure_ids:
            raise InvalidQueryError(f'There is no relative risk for the entity that you requested')
        measure_data = _pull_rr_data_from_gbd(measure_ids)
        most_detailed_causes = measure_data['cause_id'].isin([c.gbd_id for c in causes])
        measure_data = measure_data[most_detailed_causes]

    elif isinstance(entities[0], CoverageGap):
        df = []
        SPECIAL = [coverage_gaps.low_measles_vaccine_coverage_first_dose]
        special_cases = set(SPECIAL).intersection(set(entities))
        if special_cases:
            measure_data = _pull_rr_data_from_gbd([i for i in measure_ids if i is not None])
            draw_cols = [f'draw_{i}' for i in range(1000)]
            measure_data.loc[:, draw_cols] = 1 / measure_data.loc[:, draw_cols]
            measure_data = _handle_special_coverage_gap_data(special_cases, measure_data, 1)
            measure_data = measure_data.rename(columns={"rei_id": "risk_id"})
            df.append(measure_data)

        # any coverage_gap from aux_data
        if set(entities).difference(special_cases):
            for entity in entities:
                measure_data = gbd.get_auxiliary_data('relative_risk', 'coverage_gap', entity.name)
                exposure_data = gbd.get_auxiliary_data('relative_risk', 'coverage_gap', entity.name)
                missing_year_ids = set(exposure_data.year_id.unique()).difference(measure_data.year_id.unique())
                missing_data = []
                for id in missing_year_ids:
                    data = measure_data.copy()
                    data['year_id'] = id
                    missing_data = missing_data.append(data)
                measure_data = pd.concat([measure_data] + missing_data)
                measure_data['coverage_gap'] = entity.name
                measure_data = measure_data.rename(columns={"rei_id": "risk_id"})
                del measure_data['measure']
                df.append(measure_data)
        measure_data = pd.concat(df)

    measure_data = _standardize_all_age_groups(measure_data)
    measure_data = _standardize_data(measure_data, 1)
    return measure_data


def _filter_to_most_detailed(data):
    for column, entity_list in [('cause_id', causes), ('etiology_id', etiologies),
                                ('risk_id', risk_factors), ('coverage_gap_id', coverage_gaps)]:
        if column in data:
            most_detailed = {e.gbd_id for e in entity_list if e is not None}
            data = data.query(f"{column} in @most_detailed")
    return data


def _compute_paf_for_special_cases(affected_entity: Union[Cause, Risk], entity: Union[CoverageGap, Risk, Etiology],
                                   location_ids: Iterable[int]):
    """Computes pafs in cases where rr and exposure data is inconsistent with available pafs.

    e.g., Paf for unsafe_water_source from central_comp is lower than what it is
    supposed to be and does not assign correct incidence rates equivalent to gbd

    :param affected_entity: which entity is affected by this entity
    """
    ex = _get_exposure([entity], location_ids)
    rr = _get_relative_risk([entity], location_ids)

    if isinstance(entity, (Risk, Etiology)):
        cause_id, risk_id = affected_entity.gbd_id, entity.gbd_id
        rr = rr[(rr.cause_id == cause_id) & (rr.rei_id == risk_id)]
    elif isinstance(entity, CoverageGap):
        if isinstance(affected_entity, Cause):
            cause_id, risk_id = affected_entity.gbd_id, np.nan
            rr = rr[(rr.cause_id == cause_id)]
        elif isinstance(affected_entity, Risk):
            risk_id, cause_id = affected_entity.gbd_id, np.nan
            rr = rr[(rr.risk_id == risk_id)]
        else:
            raise InvalidQueryError(f'You requested the non-valid PAF data for {entity}-{affected_entity} pair')

    paf = []

    for location_id in location_ids:
        key_cols = ['age_group_id', 'year_id', 'sex_id', 'parameter']
        ex = ex[ex.location_id == location_id]
        years = rr.year_id.unique()
        relative_risk = rr.set_index(key_cols)

        exposure = ex[ex['year_id'].isin(years)].set_index(key_cols)
        draw_columns = ['draw_{}'.format(i) for i in range(1000)]
        temp = relative_risk[draw_columns]*exposure[draw_columns]
        temp_sum = temp.groupby(['age_group_id', 'year_id', 'sex_id']).sum()
        temp_result = ((temp_sum-1)/temp_sum)
        temp_result = temp_result.replace(-np.inf, 0)  # Rows with zero exposure.
        temp_result['cause_id'] = cause_id
        temp_result['location_id'] = location_id
        temp_result['risk_id'] = risk_id
        temp_result['measure_id'] = 3
        if isinstance(entity, CoverageGap):
            temp_result['coverage_gap'] = entity.name
        paf.append(temp_result.reset_index())

    paf_data = pd.concat(paf)

    if entity == coverage_gaps.low_measles_vaccine_coverage_first_dose:
        paf_data['risk_id'] = entity.gbd_id
        paf_data['coverage_gap_id'] = entity.gbd_id

    return paf_data


def _get_population_attributable_fraction(entities, location_ids):
    SPECIAL = [risk_factors.unsafe_water_source]
    measure_ids = _get_ids_for_measure(entities, 'population_attributable_fraction')

    if isinstance(entities[0], Cause):
        measure_data = gbd.get_pafs(entity_ids=measure_ids, location_ids=location_ids)
        measure_data = measure_data.rename(columns={"rei_id": "risk_id"})
        measure_data = _filter_to_most_detailed(measure_data)

        risks_in_result = measure_data.risk_id.unique()
        special_cases = [r for r in SPECIAL if r.gbd_id in risks_in_result]
        for risk in special_cases:
            special_causes = measure_data[measure_data.risk_id == risk.gbd_id].cause_id.unique()
            special_causes = [cause for cause in causes if
                              cause and cause.gbd_id in special_causes and cause is not causes.all_causes]
            for cause in special_causes:
                special_paf = _compute_paf_for_special_cases(cause, risk, location_ids)
                measure_data = measure_data.query("risk_id != @risk.gbd_id or cause_id != @cause.gbd_id")
                measure_data = measure_data.append(special_paf)

    elif isinstance(entities[0], Etiology):
        measure_data = gbd.get_pafs(entity_ids=measure_ids, location_ids=location_ids, entity_type='etiology')
        measure_data = measure_data.rename(columns={"rei_id": "etiology_id"})
        measure_data = _filter_to_most_detailed(measure_data)

    elif isinstance(entities[0], CoverageGap):
        paf = []
        for entity in entities:
            affected_risk_factors = entity.affected_risk_factors
            affected_causes = entity.affected_causes
            paf.extend([_compute_paf_for_special_cases(cause, entity, location_ids) for cause
                        in affected_causes if affected_causes])
            paf.extend([_compute_paf_for_special_cases(risk_factor, entity, location_ids) for risk_factor
                        in affected_risk_factors if affected_risk_factors])
        measure_data = pd.concat(paf)

    else:
        raise InvalidQueryError(f"Entity {entities[0].name} has no data for measure 'population_attributable_fraction'")

    # FIXME: We currently do not handle the case where PAF==1 so we just dump those rows.
    draws = [c for c in measure_data.columns if 'draw_' in c]
    measure_data = measure_data.loc[~(measure_data[draws] == 1).any(axis=1)]
    # TODO: figure out if we need to assert some property of the different PAF measures
    measure_data = measure_data[measure_data['measure_id'] == name_measure_map['YLD']]
    del measure_data['measure_id']

    # FIXME: I'm passing because this is broken for SBP, it's unimportant, and I don't have time to investigate -J.C.
    # measure_ids = {name_measure_map[m] for m in ['death', 'DALY', 'YLD', 'YLL']}
    # err_msg = ("Not all PAF data has values for deaths, DALYs, YLDs and YLLs. "
    #           + "This may not indicate an error but it is a case we don't explicitly handle. "
    #           + "If you need this PAF, come talk to one of the programmers.")
    # assert np.all(
    #    measure_data.groupby(key_columns).measure_id.unique().apply(lambda x: set(x) == measure_ids)), err_msg

    measure_data = _standardize_data(measure_data, 0)
    return measure_data


def _get_exposure(entities, location_ids):

    def handle_exposure_from_gbd(measure_data):
        is_categorical_exposure = measure_data.measure_id == name_measure_map['proportion']
        is_continuous_exposure = measure_data.measure_id == name_measure_map['continuous']
        measure_data = measure_data[is_categorical_exposure | is_continuous_exposure]
        # FIXME: The sex filtering should happen in the reshaping step.
        measure_data = measure_data[measure_data['sex_id'] != COMBINED]

        # FIXME: Is this the only data we need to delete measure id for?
        del measure_data['measure_id']
        return measure_data

    measure_ids = _get_ids_for_measure(entities, 'exposure')

    if isinstance(entities[0], (Risk, Etiology)):
        if None in measure_ids:
            raise InvalidQueryError(f'There is no exposure data for the entity that you requested')
        measure_data = gbd.get_exposures(risk_ids=measure_ids, location_ids=location_ids)
        measure_data = _handle_weird_exposure_measures(measure_data)
        exposure_data = handle_exposure_from_gbd(measure_data)

    elif isinstance(entities[0], CoverageGap):
        df = []
        SPECIAL = [coverage_gaps.low_measles_vaccine_coverage_first_dose]
        special_cases = set(SPECIAL).intersection(set(entities))
        if special_cases:
            measure_data = gbd.get_exposures(risk_ids=[s.gbd_id for s in special_cases], location_ids=location_ids)
            measure_data = _handle_special_coverage_gap_data(special_cases, measure_data, 0)
            measure_data = handle_exposure_from_gbd(measure_data)
            del measure_data['modelable_entity_id']
            del measure_data['metric_id']
            df.append(measure_data)

        # any coverage_gap from aux_data
        if set(entities).difference(special_cases):
            for entity in entities:
                measure_data = gbd.get_auxiliary_data('exposure', 'coverage_gap', entity.name)
                measure_data = measure_data[measure_data.location_id.isin(location_ids)]
                measure_data['coverage_gap'] = entity.name
                del measure_data['measure']
                df.append(measure_data)
        exposure_data = pd.concat(df)
    else:
        raise InvalidQueryError(f"Entity {entities[0].name} has no data for measure 'exposure'")

    return _standardize_all_age_groups(exposure_data)


def _handle_special_coverage_gap_data(entities, measure_data, fill_value):
    # We pulled coverage, not exposure, so invert the categories.
    coverage = measure_data['parameter'] == 'cat1'
    exposure = measure_data['parameter'] == 'cat2'
    measure_data.loc[coverage, 'parameter'] = 'cat2'
    measure_data.loc[exposure, 'parameter'] = 'cat1'
    measure_data = measure_data.rename(columns={'risk_id': 'coverage_gap_id'})

    for coverage_gap in entities:
        affected_causes = coverage_gap.affected_causes
        if len(affected_causes) != 1:
            raise UnhandledDataError("We only handle coverage gaps affecting a single cause. "
                                     "Tell James if you see this error.")
        restrictions = affected_causes[0].restrictions
        if restrictions.yll_only:
            raise UnhandledDataError("The PAFs we use are for YLDs, so causes with no attributable YLDs should not"
                                     "have associated exposures or RRs.")
        age_start, age_end = restrictions.yld_age_start, restrictions.yld_age_end
        min_age_group = age_restriction_map[age_start][0]
        max_age_group = age_restriction_map[age_end][1]

        good_age_groups = range(min_age_group, max_age_group+1)

        coverage_gap_data = measure_data['coverage_gap_id'] == coverage_gap.gbd_id
        correct_age_groups = measure_data['age_group_id'].isin(good_age_groups)
        measure_data['rei_id'] = np.NaN
        measure_data['coverage_gap'] = coverage_gap.name
        draw_cols = [f'draw_{i}' for i in range(1000)]
        measure_data.loc[coverage_gap_data & ~correct_age_groups, draw_cols] = fill_value

    return measure_data


def _handle_weird_exposure_measures(measure_data):
    key_cols = ['age_group_id', 'location_id', 'sex_id', 'year_id']
    draw_cols = [f'draw_{i}' for i in range(1000)]

    measure_data = measure_data.set_index(key_cols)
    measure_data = measure_data[draw_cols + ['risk_id', 'measure_id', 'parameter']]
    exposure_data = pd.DataFrame()

    for risk_id in measure_data.risk_id.unique():
        # We need to handle this juggling risk by risk because the data is heterogeneous by risk id.
        correct_risk = measure_data['risk_id'] == risk_id
        risk_data = measure_data[correct_risk]

        measure_id = _get_exposure_measure_id(risk_data)

        # FIXME:
        # Some categorical risks come from cause models, or they get weird exposure models that
        # report prevalence instead of proportion.  We should do a systematic review of them and work
        # with the risk factors team to get the exposure reported consistently.  In the mean time
        # we scale the unit-full prevalence numbers to unit-less proportion numbers. - J.C.
        if measure_id == name_measure_map['prevalence']:
            total_prevalence = risk_data[draw_cols].reset_index().groupby(key_cols).sum()
            for parameter in risk_data['parameter'].unique():
                correct_parameter = risk_data['parameter'] == parameter
                risk_data.loc[correct_parameter, draw_cols] /= total_prevalence
            risk_data = risk_data.reset_index()
            risk_data['measure_id'] = name_measure_map['proportion']

        exposure_data = exposure_data.append(risk_data.reset_index())
    return exposure_data


def _get_exposure_measure_id(data):
    measure_ids = data.measure_id.unique()
    if len(measure_ids) > 1:
        raise UnhandledDataError("Exposures should always come back with a single measure, "
                                 "or they should be dealt with as a special case.  ")

    return int(measure_ids)


def _get_exposure_standard_deviation(entities, location_ids):
    ids = dict(zip(_get_ids_for_measure(entities, 'exposure_standard_deviation'), [e.gbd_id for e in entities]))
    df = gbd.get_exposure_standard_deviations(list(ids.keys()), location_ids)
    key_cols = ['age_group_id', 'location_id', 'sex_id', 'year_id', 'risk_id']
    draw_cols = [f'draw_{i}' for i in range(1000)]
    df = df[df['sex_id'] != 3]
    return df[key_cols + draw_cols]


###############
# Other stuff #
###############

def _get_annual_visits(entities, location_ids):
    measure_ids = _get_ids_for_measure(entities, 'annual_visits')
    measure_data = gbd.get_modelable_entity_draws(me_ids=measure_ids, location_ids=location_ids)

    measure_data['treatment_technology'] = 'temp'
    for entity in entities:
        correct_entity = measure_data['modelable_entity_id'] == entity.utilization
        measure_data.loc[correct_entity, 'healthcare_entity'] = entity.name

    correct_measure = measure_data['measure_id'] == name_measure_map['continuous']
    correct_sex = measure_data['sex_id'] != COMBINED

    return measure_data[correct_measure & correct_sex]


def _get_cost(entities, location_ids):
    data = []
    if isinstance(entities[0], HealthcareEntity):
        for entity in entities:
            df = gbd.get_auxiliary_data('cost', 'healthcare_entity', entity.name)
            df = df[df['location_id'].isin(location_ids)]
            data.append(df)
    else:
        for entity in entities:
            df = gbd.get_auxiliary_data('cost', 'health_technology', entity.name)
            data.append(df)
    return pd.concat(data)


####################################
# Measures for risk like entities  #
####################################


def get_ensemble_weights(risks):
    data = gbd.get_auxiliary_data('ensemble_weights', 'risk_factor', 'all')
    data = data[data['risk_id'].isin([r.gbd_id for r in risks])]
    return data


#######################
# Other kinds of data #
#######################


def get_populations(locations):
    location_ids = [get_location_ids_by_name()[location] for location in locations]
    populations = pd.concat([gbd.get_populations(location_id) for location_id in location_ids])
    populations["location"] = populations.location_id.apply(get_location_names_by_id().get)
    keep_columns = ['age_group_id', 'location', 'year_id', 'sex_id', 'population']
    return populations[keep_columns]


def get_age_bins():
    return gbd.get_age_bins()


def get_theoretical_minimum_risk_life_expectancy():
    data = gbd.get_theoretical_minimum_risk_life_expectancy()
    # TODO: Figure out a more rigorous way to handle generating age bins (esp. last edge) from precise ages
    data['age_group_start'] = data['age']
    data['age_group_end'] = data.age.shift(-1).fillna(125)
    return data


def get_subregions(locations):
    location_ids = [get_location_ids_by_name()[location] for location in locations]
    return gbd.get_subregions(location_ids)


def get_covariate_estimates(covariates, locations):
    location_ids = [get_location_ids_by_name()[location] for location in locations]
    data = gbd.get_covariate_estimates([covariate.gbd_id for covariate in covariates], location_ids)
    data['location'] = data.location_id.apply(get_location_names_by_id().get)
    data = data.drop('location_id', 'columns')
    return data


def get_location_ids_by_name():
    return {r.location_name: r.location_id for _, r in gbd.get_location_ids().iterrows()}


def get_location_names_by_id():
    return {v: k for k, v in get_location_ids_by_name().items()}


def get_estimation_years():
    return gbd.get_estimation_years(gbd.GBD_ROUND_ID)




import pytest

import pandas as pd
import numpy as np

from itertools import product

from vivarium_inputs.utilities import standardize_dimensions
from vivarium_inputs.core import UnhandledDataError


def test_standardize_dimensions__fill():
    expected = pd.MultiIndex.from_product([range(22), ['Male', 'Female'], [1990, 1995, 2000, 2005]],
                                          names=['age_group_id', 'sex', 'year'])

    actual = list(product([12, 13, 14], ['Male'], [1990, 1995, 2000, 2005], ['value', 'other_value']))
    actual = pd.DataFrame(actual, columns=['age_group_id', 'sex', 'year', 'measure'])

    draws = [f'draw_{i}' for i in range(1000)]
    actual = actual.assign(**{d: 0.0 for d in draws})
    with_val = actual['measure'] == 'value'
    with_other_val = actual['measure'] == 'other_value'
    actual.loc[with_val] = actual.loc[with_val].assign(**{d: 10.0 for d in draws}, )
    actual.loc[with_other_val] = actual.loc[with_other_val].assign(**{d: 1.5 for d in draws})
    actual = actual.set_index(['age_group_id', 'sex', 'year', 'measure'])

    fills = {'value': -1, 'other_value': -2}

    new_actual = standardize_dimensions(actual.reset_index(), expected, fills)
    new_actual = new_actual.set_index(['age_group_id', 'sex', 'year'])
    actual_index = actual.reset_index().set_index(['age_group_id', 'sex', 'year']).index
    assert new_actual.index.symmetric_difference(expected).empty
    assert np.all(new_actual.loc[expected.difference(actual_index)].query('measure == "value"')[draws] == -1)
    assert np.all(new_actual.loc[expected.difference(actual_index)].query('measure == "other_value"')[draws] == -2)
    new_actual = new_actual.reset_index().set_index(['age_group_id', 'sex', 'year', 'measure'])
    assert np.all(new_actual.loc[actual.index].reindex(sorted(new_actual.columns), axis=1)
                  == actual.reindex(sorted(actual.columns), axis=1))


@pytest.mark.skip("Cluster")
def test_standardize_dimensions__interpolate():
    from core_maths.interpolate import interpolate
    expected = pd.MultiIndex.from_product([range(22), ['Male'], [1991, 1995, 2000, 2005]],
                                          names=['age_group_id', 'sex', 'year'])

    actual = list(product(range(22), ['Male'], [1990, 2005], ['value']))
    actual = pd.DataFrame(actual, columns=['age_group_id', 'sex', 'year', 'measure'])

    draws = [f'draw_{i}' for i in range(1000)]
    actual = actual.assign(**{d: 0.0 for d in draws})
    actual.loc[actual['year'] == 1990] = actual.loc[actual['year'] == 1990].assign(**{d: 10.0 for d in draws})
    actual.loc[actual['year'] == 2005] = actual.loc[actual['year'] == 2005].assign(**{d: 100.0 for d in draws})
    actual = actual.set_index(['age_group_id', 'sex', 'year', 'measure'])

    fills = {'value': -1}

    to_interpolate = pd.DataFrame({'year': [1990, 2005], 'sex': ['Male', 'Male'], 'value': [10.0, 100.0]})
    interpolated = interpolate(to_interpolate.query('year == 1990').reset_index(drop=True),
                               to_interpolate.query('year == 2005').reset_index(drop=True),
                               ['sex', 'year'], 'year', ['value'], 1990, 2005)
    interpolated = interpolated.query('year in [1991, 1995, 2000, 2005]')

    new_actual = standardize_dimensions(actual.reset_index(), expected, fills)
    new_actual = new_actual.set_index(['age_group_id', 'sex', 'year'])
    assert new_actual.index.symmetric_difference(expected).empty
    assert np.allclose(
        new_actual.reset_index()[['year', 'draw_0']].drop_duplicates('year').set_index('year').sort_index(),
        interpolated[['year', 'value']].set_index('year').sort_index()
    )


def test_standardize_dimensions__missing_dimensions():
    expected = pd.MultiIndex.from_product([range(100), ['Male', 'Female'], [1990, 1995, 2000, 2005]],
                                          names=['age_group_id', 'sex', 'year'])

    actual = list(product([80,81,82], [1990, 1995, 2000, 2005], ['value']))
    actual = pd.DataFrame(actual, columns=['age_group_id', 'year', 'measure'])

    actual['draw_0'] = 10.0
    actual = actual.set_index(['age_group_id', 'year'])
    fills = {'value': -1}

    new_actual = standardize_dimensions(actual.reset_index(), expected, fills)

    expected = expected.droplevel('sex')
    assert 'sex' not in new_actual
    new_actual = new_actual.set_index(['age_group_id', 'year'])
    assert new_actual.index.symmetric_difference(expected).empty
    assert np.all(new_actual.loc[expected.difference(actual.index)].draw_0 == -1)
    assert np.all(new_actual.loc[actual.index] == actual)


def test_standardize_dimensions__malformed():
    expected = pd.MultiIndex.from_product([range(100), ['Male', 'Female'], [1990, 1995, 2000, 2005]],
                                          names=['age_group_id', 'sex', 'year'])

    actual = list(product([10, 12, 11, 80,81,82], ['Male'], [1990, 2005], ['value']))
    actual = pd.DataFrame(actual, columns=['age_group_id', 'sex', 'year', 'measure'])

    actual['draw_0'] = 10.0
    actual = actual.set_index(['age_group_id', 'sex', 'year'])

    fills = {'value': -1}

    with pytest.raises(UnhandledDataError):
        new_actual = standardize_dimensions(actual.reset_index(), expected, fills)


def test_standardize_dimensions__missing_fills():
    expected = pd.MultiIndex.from_product([range(100), ['Male', 'Female'], [1990, 1995, 2000, 2005]],
                                          names=['age_group_id', 'sex', 'year'])

    actual = list(product([10, 12, 11, 80,81,82], ['Male'], [1990, 2005], ['value', 'other_value']))
    actual = pd.DataFrame(actual, columns=['age_group_id', 'sex', 'year', 'measure'])

    actual['draw_0'] = 10.0
    actual = actual.set_index(['age_group_id', 'sex', 'year'])

    fills = {'value': -1}

    with pytest.raises(AssertionError):
        new_actual = standardize_dimensions(actual.reset_index(), expected, fills)
